"""Bridges for PII masking task."""

import abc
from collections.abc import Callable, Iterable, Sequence
from typing import Any, Literal, TypeVar, override

import dspy
import pydantic

from sieves.data import Doc
from sieves.model_wrappers import (
    ModelType,
    ModelWrapperInferenceMode,
    dspy_,
    langchain_,
    outlines_,
)
from sieves.model_wrappers.types import ModelSettings
from sieves.tasks.predictive.bridges import Bridge
from sieves.tasks.predictive.consolidation import MultiEntityConsolidation
from sieves.tasks.predictive.schemas.pii_masking import (
    PIIEntity,
    Result,
)

_BridgePromptSignature = TypeVar("_BridgePromptSignature")
_BridgeResult = TypeVar("_BridgeResult")


class PIIMaskingBridge(Bridge[_BridgePromptSignature, _BridgeResult, ModelWrapperInferenceMode], abc.ABC):
    """Abstract base class for PII masking bridges."""

    def __init__(
        self,
        task_id: str,
        prompt_instructions: str | None,
        mask_placeholder: str,
        pii_types: Sequence[str] | dict[str, str] | None,
        overwrite: bool,
        model_settings: ModelSettings,
        prompt_signature: type[pydantic.BaseModel],
        model_type: ModelType,
        fewshot_examples: Sequence[pydantic.BaseModel] = (),
    ):
        """Initialize PII masking bridge.

        :param task_id: Task ID.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param mask_placeholder: Placeholder for masked PII.
        :param pii_types: PII types to mask.
        :param overwrite: Whether to overwrite original text.
        :param model_settings: Settings for structured generation.
        :param prompt_signature: Unified Pydantic prompt signature.
        :param model_type: Model type.
        :param fewshot_examples: Few-shot examples.
        """
        super().__init__(
            task_id=task_id,
            prompt_instructions=prompt_instructions,
            overwrite=overwrite,
            model_settings=model_settings,
            prompt_signature=prompt_signature,
            model_type=model_type,
            fewshot_examples=fewshot_examples,
        )

        self._mask_placeholder = mask_placeholder
        self._pii_types: list[str] | None = None
        self._pii_type_descriptions: dict[str, str] = {}

        if isinstance(pii_types, dict):
            self._pii_types = list(pii_types.keys())
            self._pii_type_descriptions = pii_types
        elif pii_types is not None:
            self._pii_types = pii_types
            self._pii_type_descriptions = {}

        self._pii_entity_cls = self._create_pii_entity_cls()
        self._consolidation_strategy = MultiEntityConsolidation(extractor=self._chunk_extractor)

    @property
    @abc.abstractmethod
    def _chunk_extractor(self) -> Callable[[Any], Iterable[pydantic.BaseModel]]:
        """Return a callable that extracts a list of entities from a raw chunk result.

        :return: Extractor callable.
        """

    def _get_pii_type_descriptions(self) -> str:
        """Return a string with the PII type descriptions.

        :return: A string with the PII type descriptions.
        """
        if not self._pii_types:
            return ""

        pii_types_with_descriptions: list[str] = []
        for pii_type in self._pii_types:
            if pii_type in self._pii_type_descriptions:
                pii_types_with_descriptions.append(
                    f"  <pii_type_description>\n    <pii_type>{pii_type}</pii_type>\n    <description>"
                    f"{self._pii_type_descriptions[pii_type]}</description>\n  </pii_type_description>"
                )
            else:
                pii_types_with_descriptions.append(f"  <pii_type>{pii_type}</pii_type>")

        pii_type_desc_string = "\n".join(pii_types_with_descriptions)
        return f"<pii_type_descriptions>\n{pii_type_desc_string}\n</pii_type_descriptions>"

    def _create_pii_entity_cls(self) -> type[pydantic.BaseModel]:
        """Create PII entity class.

        :returns: PII entity class.
        """
        pii_types_list = []
        if self._pii_types:
            for pt in self._pii_types:
                pii_types_list.append(pt)
                if pt.lower() not in pii_types_list:
                    pii_types_list.append(pt.lower())
                if pt.upper() not in pii_types_list:
                    pii_types_list.append(pt.upper())

        PIIType = Literal[*pii_types_list] if pii_types_list else str  # type: ignore[invalid-type-form]

        class PIIEntityRuntime(PIIEntity, frozen=True):
            """PII entity."""

            entity_type: PIIType  # type: ignore[valid-type]

        return PIIEntityRuntime


class DSPyPIIMasking(PIIMaskingBridge[dspy_.PromptSignature, dspy_.Result, dspy_.InferenceMode]):
    """DSPy bridge for PII masking."""

    @override
    def _validate(self) -> None:
        assert self._model_type == ModelType.dspy

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        return ""

    @override
    @property
    def inference_mode(self) -> dspy_.InferenceMode:
        """Return inference mode for DSPy model wrapper."""
        return self._model_settings.inference_mode or dspy_.InferenceMode.predict

    @property
    @override
    def _chunk_extractor(self) -> Callable[[Any], Iterable[pydantic.BaseModel]]:
        return lambda res: res.pii_entities

    @override
    def integrate(self, results: Sequence[dspy_.Result], docs: list[Doc]) -> list[Doc]:
        """Integrate results into docs."""
        for doc, result in zip(docs, results):
            # Store masked text and PII entities in results
            res = Result(
                masked_text=result.masked_text,
                pii_entities=[
                    PIIEntity.model_validate(e.model_dump() if hasattr(e, "model_dump") else e)
                    for e in result.pii_entities
                ],
            )
            doc.results[self._task_id] = res

            if self._overwrite:
                doc.text = result.masked_text

        return docs

    @override
    def consolidate(
        self,
        results: Sequence[dspy_.Result],
        docs_offsets: list[tuple[int, int]],
    ) -> Sequence[dspy_.Result]:
        """Consolidate results from multiple chunks."""
        # Delegate consolidation of entities to strategy.
        consolidated_entities_all = self._consolidation_strategy.consolidate(results, docs_offsets)

        # Merge results for each document.
        consolidated_results: list[dspy_.Result] = []
        for i, (start, end) in enumerate(docs_offsets):
            doc_results = results[start:end]
            masked_texts: list[str] = []

            for res in doc_results:
                masked_texts.append(res.masked_text)

            consolidated_results.append(
                dspy.Prediction.from_completions(
                    {
                        "masked_text": [" ".join(masked_texts).strip()],
                        "pii_entities": [consolidated_entities_all[i]],
                    },
                    signature=self.prompt_signature,
                )
            )
        return consolidated_results


class PydanticPIIMasking(PIIMaskingBridge[pydantic.BaseModel, pydantic.BaseModel, ModelWrapperInferenceMode], abc.ABC):
    """Base class for Pydantic-based PII masking bridges."""

    @override
    def _validate(self) -> None:
        assert self._model_type in {ModelType.langchain, ModelType.outlines}

    @property
    @override
    def _chunk_extractor(self) -> Callable[[Any], Iterable[pydantic.BaseModel]]:
        return lambda res: res.pii_entities

    @property
    def _default_prompt_instructions(self) -> str:
        pii_type_info = self._get_pii_type_descriptions() if self._pii_type_descriptions else ""
        return (
            "Identify and mask Personally Identifiable Information (PII) in the given text.\n"
            "{%- if pii_types|length > 0 %}\n"
            "Focus on these specific PII types: {{ pii_types|join(', ') }}.\n"
            "{%- else %}\n"
            "Mask all common types of PII such as names, addresses, phone numbers, emails, SSNs, credit card numbers, "
            "etc.\n"
            "{%- endif %}\n"
            f"{pii_type_info}\n"
            f'Replace each instance of PII with "{self._mask_placeholder}".\n'
            "Provide a confidence score between 0.0 and 1.0 for each entity found."
        )

    @override
    @property
    def _prompt_conclusion(self) -> str | None:
        return "========\n\n<text>{{ text }}</text>"

    @override
    def integrate(self, results: Sequence[pydantic.BaseModel], docs: list[Doc]) -> list[Doc]:
        for doc, result in zip(docs, results):
            assert hasattr(result, "masked_text")
            assert hasattr(result, "pii_entities")
            # Store masked text and PII entities in results
            doc.results[self._task_id] = Result(
                masked_text=result.masked_text,
                pii_entities=[PIIEntity.model_validate(e) for e in result.pii_entities],
            )

            if self._overwrite:
                doc.text = result.masked_text

        return docs

    @override
    def consolidate(
        self,
        results: Sequence[pydantic.BaseModel],
        docs_offsets: list[tuple[int, int]],
    ) -> Sequence[pydantic.BaseModel]:
        assert issubclass(self.prompt_signature, pydantic.BaseModel)

        consolidated_entities_all = self._consolidation_strategy.consolidate(results, docs_offsets)
        consolidated_results: list[pydantic.BaseModel] = []

        for i, (start, end) in enumerate(docs_offsets):
            doc_results = results[start:end]
            masked_texts: list[str] = []

            for res in doc_results:
                if res is None:
                    continue  # type: ignore[unreachable]

                assert hasattr(res, "masked_text")
                masked_texts.append(res.masked_text)

            consolidated_results.append(
                self.prompt_signature(
                    masked_text=" ".join(masked_texts).strip(),
                    pii_entities=consolidated_entities_all[i],
                )
            )

        return consolidated_results

    @override
    @property
    def model_type(self) -> ModelType:
        return self._model_type

    @override
    @property
    def inference_mode(self) -> outlines_.InferenceMode | langchain_.InferenceMode:
        if self._model_type == ModelType.outlines:
            return self._model_settings.inference_mode or outlines_.InferenceMode.json
        elif self._model_type == ModelType.langchain:
            return self._model_settings.inference_mode or langchain_.InferenceMode.structured

        raise ValueError(f"Unsupported model type: {self._model_type}")



"""Schemas for sentiment analysis task."""

from __future__ import annotations

import dspy
import pydantic

from sieves.model_wrappers import dspy_, langchain_, outlines_
from sieves.tasks.predictive.schemas.core import FewshotExample as BaseFewshotExample


class FewshotExample(BaseFewshotExample):
    """Example for sentiment analysis few-shot prompting.

    Attributes:
        text: Input text.
        sentiment_per_aspect: Mapping of aspects to sentiments.
        score: Confidence score for the sentiment assessment.
    """

    sentiment_per_aspect: dict[str, float]
    score: float | None = None

    @property
    def target_fields(self) -> tuple[str, ...]:
        """Return target fields.

        :return: Target fields.
        """
        return ("sentiment_per_aspect", "score")


# --8<-- [start:Result]
class Result(pydantic.BaseModel):
    """Result of an aspect-based sentiment analysis task.

    Attributes:
        sentiment_per_aspect: Mapping of aspects to sentiment scores.
        score: Overall confidence score for the sentiment analysis.
    """

    sentiment_per_aspect: dict[str, float] = pydantic.Field(
        description="A mapping where keys are aspect names (e.g., 'overall', 'quality') and values are sentiment "
        "scores between 0 (Negative) and 1 (Positive)."
    )
    score: float | None = pydantic.Field(
        default=None, description="Provide an overall confidence score for the sentiment analysis, between 0 and 1."
    )


# --8<-- [end:Result]


TaskModel = dspy_.Model | langchain_.Model | outlines_.Model
TaskPromptSignature = type[dspy.Signature] | type[pydantic.BaseModel]
TaskResult = Result



"""Schemas for question answering task."""

from __future__ import annotations

import dspy
import pydantic

from sieves.model_wrappers import dspy_, langchain_, outlines_
from sieves.tasks.predictive.schemas.core import FewshotExample as BaseFewshotExample


class QuestionAnswer(pydantic.BaseModel):
    """A pair consisting of a question, its generated answer, and a confidence score.

    Attributes:
        question: Question asked.
        answer: Answer to the question.
        score: Confidence score.
    """

    question: str = pydantic.Field(
        description="The question being asked. This should exactly match one of the provided questions."
    )
    answer: str = pydantic.Field(description="The generated answer to the question based on the document.")
    score: float | None = pydantic.Field(
        default=None, description="Provide a confidence score for the generated answer, between 0 and 1."
    )


class FewshotExample(BaseFewshotExample):
    """Few-shot example with questions and answers for a context.

    Attributes:
        text: Input text.
        questions: Questions asked.
        answers: Expected answers.
        scores: Confidence scores for answers.
    """

    text: str
    questions: list[str]
    answers: list[str]
    scores: list[float] | None = None

    @property
    def input_fields(self) -> tuple[str, ...]:
        """Return input fields.

        :return: Input fields.
        """
        return ("text", "questions")

    @property
    def target_fields(self) -> tuple[str, ...]:
        """Return target fields.

        :return: Target fields.
        """
        return ("answers", "scores")

    def to_dspy(self) -> dspy.Example:
        """Convert to `dspy.Example` with qa_pairs.

        :returns: Example as `dspy.Example`.
        """
        scores = self.scores or [None] * len(self.answers)
        qa_pairs = [
            QuestionAnswer(
                question=q,
                answer=a,
                score=s,
            )
            for q, a, s in zip(self.questions, self.answers, scores)
        ]

        return dspy.Example(text=self.text, questions=self.questions, qa_pairs=qa_pairs).with_inputs(*self.input_fields)


# --8<-- [start:Result]
class Result(pydantic.BaseModel):
    """Result of a question-answering task. Contains a list of question-answer pairs.

    Attributes:
        qa_pairs: List of question-answer pairs.
    """

    qa_pairs: list[QuestionAnswer] = pydantic.Field(
        description="A list of question-answer pairs corresponding to the provided questions."
    )


# --8<-- [end:Result]


TaskModel = dspy_.Model | langchain_.Model | outlines_.Model
TaskPromptSignature = type[dspy.Signature] | type[pydantic.BaseModel]
TaskResult = Result



"""
Test file containing examples for the Distillation guide.

These code blocks are referenced in docs/guides/distillation.md using snippet injection.

IMPORTANT: These tests use MINIMAL settings to reduce cost and time:
- num_epochs=1 (instead of default 3-5)
- Only 9 training examples (instead of 100+)
- Minimal batch sizes
"""

import pytest
from tempfile import TemporaryDirectory
from pathlib import Path


def test_basic_setfit_distillation(small_dspy_model):
    """Test basic SetFit distillation with minimal settings."""
    model = small_dspy_model

    # --8<-- [start:distillation-setfit-basic]
    # --8<-- [start:distillation-setfit-imports]
    import dspy
    from sieves import tasks, Doc, Pipeline
    from sieves.tasks import DistillationFramework
    from setfit import SetFitModel
    # --8<-- [end:distillation-setfit-imports]

    # --8<-- [start:distillation-setfit-data]
    # 1. Create MINIMAL training data (9 examples - 3 per label)
    docs = [
        Doc(text="New AI model released", meta={"label": "technology"}),
        Doc(text="Software update announced", meta={"label": "technology"}),
        Doc(text="Smartphone innovation unveiled", meta={"label": "technology"}),
        Doc(text="Election results in", meta={"label": "politics"}),
        Doc(text="Parliament votes on bill", meta={"label": "politics"}),
        Doc(text="Political debate tonight", meta={"label": "politics"}),
        Doc(text="Team wins game", meta={"label": "sports"}),
        Doc(text="Championship final scores", meta={"label": "sports"}),
        Doc(text="Olympic athlete breaks record", meta={"label": "sports"}),
    ]
    # --8<-- [end:distillation-setfit-data]

    # --8<-- [start:distillation-setfit-teacher]
    # 2. Define teacher task
    teacher = tasks.Classification(
        labels=["technology", "politics", "sports"],
        model=model,
    )

    # 3. Process documents to generate labels
    pipeline = Pipeline([teacher])
    labeled_docs = list(pipeline(docs))
    # --8<-- [end:distillation-setfit-teacher]

    # --8<-- [start:distillation-setfit-distill]
    # 4. Distill with MINIMAL training settings
    with TemporaryDirectory() as tmp_dir:
        teacher.distill(
            base_model_id="sentence-transformers/paraphrase-albert-small-v2",
            framework=DistillationFramework.setfit,
            data=labeled_docs,
            output_path=tmp_dir,
            val_frac=0.3,
            seed=42,
            train_kwargs={
                "num_iterations": 1,    # Min. iterations.
                "num_epochs": 1,        # Minimal epochs (instead of 3-5)
                "batch_size": 8,
            },
        )
        # --8<-- [end:distillation-setfit-distill]

        # --8<-- [start:distillation-setfit-load]
        # 5. Load and use distilled model
        distilled_model = SetFitModel.from_pretrained(tmp_dir)
        predictions = distilled_model.predict(["Smartphone announcement"])
        print(f"Predictions: {predictions}")
        # --8<-- [end:distillation-setfit-load]
    # --8<-- [end:distillation-setfit-basic]

        assert predictions is not None


def test_to_hf_dataset_export(small_dspy_model):
    """Test exporting task results to HuggingFace dataset."""
    model = small_dspy_model

    # --8<-- [start:distillation-to-hf-dataset]
    import dspy
    from sieves import tasks, Doc, Pipeline

    # 1. Create minimal training data
    docs = [
        Doc(text="AI breakthrough announced", meta={"label": "technology"}),
        Doc(text="Election campaign begins", meta={"label": "politics"}),
        Doc(text="Sports team victory", meta={"label": "sports"}),
    ]

    # 2. Run classification task
    task = tasks.Classification(
        labels=["technology", "politics", "sports"],
        model=model,
    )

    pipeline = Pipeline([task])
    labeled_docs = list(pipeline(docs))

    # 3. Export to HuggingFace dataset
    hf_dataset = task.to_hf_dataset(labeled_docs)

    print(f"Dataset features: {hf_dataset.features}")
    print(f"Dataset size: {len(hf_dataset)}")
    # --8<-- [end:distillation-to-hf-dataset]

    assert len(hf_dataset) == 3
    assert "text" in hf_dataset.features
    assert "labels" in hf_dataset.features



# mypy: ignore-errors
import pytest

from sieves import Doc, Pipeline
from sieves.model_wrappers import ModelType, ModelSettings, dspy_, langchain_, outlines_
from sieves.serialization import Config
from sieves.tasks import PredictiveTask
from sieves.tasks.predictive import translation
from sieves.tasks.predictive.schemas.translation import Result


@pytest.mark.parametrize(
    "batch_runtime",
    translation.Translation.supports(),
    indirect=["batch_runtime"],
)
@pytest.mark.parametrize("fewshot", [True, False])
def test_run(translation_docs, batch_runtime, fewshot) -> None:
    fewshot_examples = [
        translation.FewshotExample(
            text="The sun is shining today.",
            to="Spanish",
            translation="El sol brilla hoy.",
            score=1.0,
        ),
        translation.FewshotExample(
            text="There's a lot of fog today",
            to="Spanish",
            translation="Hay mucha niebla hoy.",
            score=1.0,
        ),
    ]

    fewshot_args = {"fewshot_examples": fewshot_examples} if fewshot else {}
    task = translation.Translation(
        to="Spanish",
        model=batch_runtime.model,
        model_settings=batch_runtime.model_settings,
        batch_size=batch_runtime.batch_size,
        **fewshot_args,
    )
    pipe = Pipeline(task)
    docs = list(pipe(translation_docs))

    assert len(docs) == 2
    for doc in docs:
        assert doc.text
        assert "Translation" in doc.results

        # Verify unified result types.
        assert isinstance(doc.results["Translation"], translation.Result)

        assert "Translation" in doc.meta
        assert "raw" in doc.meta["Translation"]
        assert "usage" in doc.meta["Translation"]
        assert "usage" in doc.meta

        print(f"Output: {doc.results['Translation']}")
        print(f"Raw output: {doc.meta['Translation']['raw']}")
        print(f"Usage: {doc.meta['Translation']['usage']}")
        print(f"Total Usage: {doc.meta['usage']}")

    with pytest.raises(NotImplementedError):
        pipe["Translation"].distill(None, None, None, None, None, None, None, None)

    if fewshot:
        _to_hf_dataset(task, docs)


def _to_hf_dataset(task: translation.Translation, docs: list[Doc]) -> None:
    """Tests whether conversion to HF dataset works as expected.

    :param task: Translation task instance.
    :param docs: List of documents to convert.
    """
    dataset = task.to_hf_dataset(docs)
    assert all([key in dataset.features for key in ("text", "translation")])
    assert len(dataset) == 2
    records = list(dataset)
    assert records[0]["text"] == "It is rainy today."
    assert records[1]["text"] == "It is cloudy today."
    for record in records:
        assert isinstance(record["translation"], str)

    with pytest.raises(KeyError):
        task.to_hf_dataset([Doc(text="This is a dummy text.")])


@pytest.mark.parametrize("batch_runtime", [ModelType.dspy], indirect=["batch_runtime"])
def test_serialization(translation_docs, batch_runtime) -> None:
    pipe = Pipeline([
        translation.Translation(
            to="Spanish",
            model=batch_runtime.model,
            model_settings=batch_runtime.model_settings,
            batch_size=batch_runtime.batch_size,
        )
    ])

    config = pipe.serialize()
    assert config.model_dump() == {'cls_name': 'sieves.pipeline.core.Pipeline',
 'tasks': {'is_placeholder': False,
           'value': [{'cls_name': 'sieves.tasks.predictive.translation.core.Translation',
                      'fewshot_examples': {'is_placeholder': False,
                                           'value': ()},
                      'batch_size': {'is_placeholder': False, "value": -1},
                      'model_settings': {'is_placeholder': False,
                                              'value': {
                                                        'config_kwargs': None,
                                                        'inference_kwargs': None,
                                                        'init_kwargs': None,
                                                        'strict': True, 'inference_mode': None}},
                      'include_meta': {'is_placeholder': False, 'value': True},
                      'model': {'is_placeholder': True,
                                'value': 'dspy.clients.lm.LM'},
                      'prompt_instructions': {'is_placeholder': False,
                                          'value': None},
                      'task_id': {'is_placeholder': False,
                                  'value': 'Translation'},
                      'condition': {'is_placeholder': False, 'value': None},
                      'to': {'is_placeholder': False, 'value': 'Spanish'},
                      'version': Config.get_version()}]},
 'use_cache': {'is_placeholder': False, 'value': True},
 'version': Config.get_version()}

    Pipeline.deserialize(config=config, tasks_kwargs=[{"model": batch_runtime.model}])


@pytest.mark.parametrize(
    "batch_runtime",
    translation.Translation.supports(),
    indirect=["batch_runtime"],
)
def test_inference_mode_override(batch_runtime) -> None:
    """Test that inference_mode parameter overrides the default value."""
    dummy = "dummy_inference_mode"

    task = translation.Translation(
        to="Spanish",
        model=batch_runtime.model,
        model_settings=ModelSettings(inference_mode=dummy),
        batch_size=batch_runtime.batch_size,
    )

    assert task._bridge.inference_mode == dummy


@pytest.mark.parametrize("batch_runtime", [ModelType.dspy], indirect=["batch_runtime"])
def test_evaluation(batch_runtime) -> None:
    """Test evaluation for translation using a real judge."""
    task = translation.Translation(to="Spanish", model=batch_runtime.model, task_id="trans")

    # 1. Full overlap
    doc_full = Doc(text="Hello world")
    res_full = Result(translation="Hola mundo", score=1.0)
    doc_full.results["trans"] = res_full
    doc_full.gold["trans"] = res_full
    report_full = task.evaluate([doc_full], judge=batch_runtime.model)
    assert report_full.metrics[task.metric] > 0.8

    # 2. No overlap
    doc_none = Doc(text="Hello world")
    doc_none.results["trans"] = Result(translation="Adiós mundo", score=1.0)
    doc_none.gold["trans"] = Result(translation="Hola mundo", score=1.0)
    report_none = task.evaluate([doc_none], judge=batch_runtime.model)
    assert report_none.metrics[task.metric] < 0.6

    # 3. Partial overlap
    doc_partial = Doc(text="Hello world")
    doc_partial.results["trans"] = Result(translation="Hola", score=1.0)
    doc_partial.gold["trans"] = Result(translation="Hola mundo", score=1.0)
    report_partial = task.evaluate([doc_partial], judge=batch_runtime.model)
    assert 0.3 < report_partial.metrics[task.metric] < 0.8



"""Schemas for summarization task."""

from __future__ import annotations

from collections.abc import Sequence

import dspy
import pydantic

from sieves.model_wrappers import dspy_, langchain_, outlines_
from sieves.tasks.predictive.schemas.core import FewshotExample as BaseFewshotExample


class FewshotExample(BaseFewshotExample):
    """Example for summarization few-shot prompting.

    Attributes:
        text: Input text.
        n_words: Guideline for summary length.
        summary: Summary of text.
        score: Confidence score.
    """

    n_words: int
    summary: str
    score: float | None = None

    @property
    def input_fields(self) -> Sequence[str]:
        """Return input fields.

        :return: Input fields.
        """
        return "text", "n_words"

    @property
    def target_fields(self) -> tuple[str, ...]:
        """Return target fields.

        :return: Target fields.
        """
        return ("summary",)


# --8<-- [start:Result]
class Result(pydantic.BaseModel):
    """Result of a summarization task. Contains the generated summary and a confidence score.

    Attributes:
        summary: Summary of text.
        score: Confidence score.
    """

    summary: str = pydantic.Field(description="The generated summary of the input text.")
    score: float | None = pydantic.Field(
        default=None, description="Provide a confidence score for the generated summary, between 0 and 1."
    )


# --8<-- [end:Result]


TaskModel = dspy_.Model | langchain_.Model | outlines_.Model
TaskPromptSignature = type[dspy.Signature] | type[pydantic.BaseModel]
TaskResult = Result



"""Translation task."""

from sieves.tasks.predictive.schemas.translation import (
    FewshotExample,
    Result,
    TaskPromptSignature,
    TaskResult,
)
from sieves.tasks.predictive.translation.core import Translation

__all__ = ["Translation", "FewshotExample", "Result", "TaskResult", "TaskPromptSignature"]



# mypy: ignore-errors
import pytest

from sieves import Pipeline
from sieves.model_wrappers import ModelType
from sieves.serialization import Config
from sieves.tasks.preprocessing.chunking.naive import NaiveChunker


@pytest.mark.parametrize(
    "batch_runtime",
    [ModelType.huggingface],
    indirect=["batch_runtime"],
)
def test_run(dummy_docs, batch_runtime) -> None:
    """Tests whether chunking mechanism in PredictiveTask works as expected."""
    chunk_interval = 5
    pipe = Pipeline([NaiveChunker(interval=chunk_interval)])
    docs = list(pipe(dummy_docs))

    assert len(docs) == 2
    for doc in docs:
        assert doc.text
        assert len(doc.chunks) == 2


def test_serialization(dummy_docs) -> None:
    chunk_interval = 5
    pipe = Pipeline(tasks=[NaiveChunker(interval=chunk_interval)])
    docs = list(pipe(dummy_docs))

    config = pipe.serialize()
    assert config.model_dump() == {
        "cls_name": "sieves.pipeline.core.Pipeline",
        "use_cache": {"is_placeholder": False, "value": True},
        "tasks": {
            "is_placeholder": False,
            "value": [
                {
                    "cls_name": "sieves.tasks.preprocessing.chunking.naive.NaiveChunker",
                    'batch_size': {'is_placeholder': False, "value": -1},
                    "include_meta": {"is_placeholder": False, "value": False},
                    "interval": {"is_placeholder": False, "value": 5},
                    "task_id": {"is_placeholder": False, "value": "NaiveChunker"},
                    "version": Config.get_version(),
                    'condition': {'is_placeholder': False, 'value': None},
                }
            ],
        },
        "version": Config.get_version(),
    }

    deserialized_pipeline = Pipeline.deserialize(config=config, tasks_kwargs=[{}])
    assert docs[0] == list(deserialized_pipeline(dummy_docs))[0]



"""Schemas for NER task."""

from __future__ import annotations

import dspy
import gliner2.inference.engine
import pydantic

from sieves.model_wrappers import dspy_, gliner_, langchain_, outlines_
from sieves.tasks.predictive.schemas.core import FewshotExample as BaseFewshotExample


class EntityWithContext(pydantic.BaseModel):
    """Entity mention with its text span, context, and type.

    Attributes:
        text: The specific text segment identified as an entity.
        context: The surrounding text providing context for the entity.
        entity_type: The category or type of the entity.
        score: Confidence score for the entity identification.
    """

    text: str = pydantic.Field(description="The specific text segment identified as an entity.")
    context: str = pydantic.Field(description="The surrounding text providing context for the entity.")
    entity_type: str = pydantic.Field(description="The category or type of the entity (e.g., PERSON, ORGANIZATION).")
    score: float | None = pydantic.Field(
        default=None, description="Provide a confidence score for the entity identification, between 0 and 1."
    )


class Entity(pydantic.BaseModel):
    """Class for storing entity information.

    Attributes:
        text: Entity text.
        start: Start offset.
        end: End offset.
        entity_type: Type of entity.
        score: Confidence score.
    """

    text: str
    start: int
    end: int
    entity_type: str
    score: float | None = None

    def __eq__(self, other: object) -> bool:
        """Compare two entities.

        :param other: Other entity to compare with.
        :return: True if entities are equal, False otherwise.
        """
        if not isinstance(other, Entity):
            return False
        return (
            self.start == other.start
            and self.end == other.end
            and self.text == other.text
            and self.entity_type == other.entity_type
        )

    def __hash__(self) -> int:
        """Compute entity hash.

        :returns: Entity hash.
        """
        return hash((self.start, self.end, self.text, self.entity_type))


class Entities(pydantic.BaseModel):
    """Collection of entities with associated text.

    Attributes:
        entities: List of entities.
        text: Source text.
    """

    entities: list[Entity]
    text: str


# --8<-- [start:Result]
class Result(Entities):
    """Result of a named-entity recognition (NER) task. Contains the extracted entities and the source text."""

    pass


# --8<-- [end:Result]


class FewshotExample(BaseFewshotExample):
    """Few‑shot example with entities annotated in text.

    Attributes:
        text: Input text.
        entities: List of entities with context.
    """

    text: str
    entities: list[EntityWithContext]

    @property
    def target_fields(self) -> tuple[str, ...]:
        """Return target fields.

        :return: Target fields.
        """
        return ("entities",)


TaskModel = dspy_.Model | gliner_.Model | langchain_.Model | outlines_.Model
TaskPromptSignature = (
    type[dspy.Signature]
    | type[pydantic.BaseModel]
    | gliner2.inference.engine.Schema
    | gliner2.inference.engine.StructureBuilder
)
TaskResult = Result



"""
Test file containing examples for the Pipeline documentation.

These code blocks are referenced in docs/pipeline.md using snippet injection.

Usage in markdown:
    ```python
    --8<-- "sieves/tests/docs/test_pipeline_docs.py:pipeline-creation-patterns"
    ```
"""


def test_pipeline_creation_patterns(example_chunker, small_transformer_model):
    """Test various ways to create and combine pipelines."""
    chunker = example_chunker
    model = small_transformer_model

    # --8<-- [start:pipeline-creation-patterns]
    from sieves import Pipeline, tasks

    # Verbose initialization (allows non-default configuration).
    t_ingest = tasks.preprocessing.Ingestion(export_format="markdown")
    t_chunk = tasks.preprocessing.Chunking(chunker)
    t_cls = tasks.predictive.Classification(labels=["science", "politics"], model=model)
    pipe = Pipeline([t_ingest, t_chunk, t_cls], use_cache=True)

    # Succinct chaining (equivalent task order).
    pipe2 = t_ingest + t_chunk + t_cls

    # You can also chain pipelines and tasks.
    pipe_left = Pipeline(t_ingest)
    pipe_right = Pipeline([t_chunk, t_cls])
    pipe3 = pipe_left + pipe_right  # results in [t_ingest, t_chunk, t_cls]

    # In-place append (mutates the left pipeline).
    # Create new pipeline for demonstration
    pipe_mutable = Pipeline([t_ingest])
    pipe_mutable += t_chunk  # appends t_chunk

    # Can also append entire pipelines
    pipe_to_append = Pipeline([tasks.predictive.Classification(labels=["tech", "sports"], model=model)])
    pipe_mutable += pipe_to_append  # appends all tasks from pipe_to_append

    # Note:
    # - Additional Pipeline parameters (e.g., use_cache=False) are only settable via the verbose form
    # - Chaining never mutates existing tasks or pipelines; it creates a new Pipeline
    # - Using "+=" mutates the existing pipeline by appending tasks
    # --8<-- [end:pipeline-creation-patterns]

    # Assertions for testing
    assert pipe is not None
    assert len(pipe.tasks) == 3


def test_conditional_execution_basic(small_transformer_model):
    """Test basic conditional task execution."""
    model = small_transformer_model

    # --8<-- [start:conditional-execution-basic]
    from sieves import Pipeline, tasks, Doc

    docs = [
        Doc(text="short"),
        Doc(text="this is a much longer document that will be processed"),
        Doc(text="med"),
    ]

    # Define a condition function
    def is_long(doc: Doc) -> bool:
        return len(doc.text or "") > 20

    # Create a task with a condition
    task = tasks.Classification(
        labels=["science", "politics"],
        model=model,
        condition=is_long
    )

    # Run pipeline
    pipe = Pipeline([task])
    for doc in pipe(docs):
        # doc.results[task.id] will be None for documents that failed the condition
        print(doc.results[task.id])
    # --8<-- [end:conditional-execution-basic]

    # Assertions for testing
    results = list(pipe(docs))
    assert len(results) == 3
    # First and third docs should be skipped (too short)
    assert results[0].results[task.id] is None
    assert results[2].results[task.id] is None
    # Second doc should be processed (long enough)
    assert results[1].results[task.id] is not None


def test_conditional_execution_multiple(example_chunker, small_transformer_model):
    """Test multiple tasks with different conditions."""
    chunker = example_chunker
    model = small_transformer_model

    # --8<-- [start:conditional-execution-multiple]
    from sieves import Pipeline, tasks, Doc

    docs = [
        Doc(text="short"),
        Doc(text="this is a much longer document"),
        Doc(text="medium text here"),
    ]

    # Task 1: Process only documents longer than 10 characters
    task1 = tasks.Chunking(chunker, condition=lambda d: len(d.text or "") > 10)

    # Task 2: Process only documents longer than 20 characters
    task2 = tasks.Classification(
        labels=["science", "politics"],
        model=model,
        condition=lambda d: len(d.text or "") > 20
    )

    # First doc: skipped by both tasks (too short)
    # Second doc: processed by both tasks (long enough)
    # Third doc: processed by task1, skipped by task2
    pipe = Pipeline([task1, task2])
    for doc in pipe(docs):
        print(doc.chunks, doc.results[task2.id])
    # --8<-- [end:conditional-execution-multiple]

    # Assertions for testing
    results = list(pipe(docs))
    assert len(results) == 3
    # First doc: skipped by both (condition failed)
    assert results[0].results[task2.id] is None
    # Second doc: processed by both tasks (long enough)
    assert results[1].results[task2.id] is not None
    # Third doc: processed by task1, skipped by task2
    assert results[2].results[task2.id] is None



"""Consolidation strategies for predictive tasks."""

from __future__ import annotations

import abc
from collections import Counter, defaultdict
from collections.abc import Callable, Iterable, Sequence
from typing import Any, Literal, TypeVar

import pydantic

_EntityType = TypeVar("_EntityType", bound=pydantic.BaseModel)


def _average_scores(scores: list[float]) -> float | None:
    """Calculate the average of a list of scores.

    :param scores: List of float scores.
    :return: Average score or None if list is empty.
    """
    return sum(scores) / len(scores) if scores else None


def _get_entity_key(entity: pydantic.BaseModel) -> str:
    """Generate a hashable key for a Pydantic model by excluding the 'score' field.

    :param entity: The Pydantic model to key.
    :return: JSON string representation of the model without its score.
    """
    return entity.model_copy(update={"score": None}).model_dump_json()


class ConsolidationStrategy(abc.ABC):
    """Abstract base class for consolidation strategies."""

    @abc.abstractmethod
    def consolidate(self, results: Sequence[Any], docs_offsets: list[tuple[int, int]]) -> Sequence[Any]:
        """Consolidate chunk results into document results.

        :param results: Sequence of raw chunk results.
        :param docs_offsets: List of (start, end) offsets mapping chunks to documents.
        :return: Sequence of consolidated "clean" results.
        """


class MultiEntityConsolidation(ConsolidationStrategy):
    """Consolidation strategy for multiple entities."""

    def __init__(self, extractor: Callable[[Any], Iterable[pydantic.BaseModel]]):
        """Initialize MultiEntityConsolidation.

        :param extractor: Callable to extract a list of entities from a raw chunk result.
        """
        self._extractor = extractor

    def consolidate(
        self,
        results: Sequence[Any],
        docs_offsets: list[tuple[int, int]],
    ) -> Sequence[list[pydantic.BaseModel]]:
        """Consolidate multiple entities from chunks.

        :param results: Raw chunk results.
        :param docs_offsets: Chunk offsets per document.
        :return: Consolidated list of entities per document.
        """
        consolidated_results: list[list[pydantic.BaseModel]] = []
        for start, end in docs_offsets:
            entities = [e for res in results[start:end] if res is not None for e in self._extractor(res)]
            consolidated_results.append(self._consolidate_entities(entities))
        return consolidated_results

    @staticmethod
    def _consolidate_entities(entities: list[_EntityType]) -> list[_EntityType]:
        """Deduplicate entities and average their scores."""
        if not entities:
            return []

        entities_map: dict[str, tuple[_EntityType, list[float]]] = {}
        for entity in entities:
            if entity is None:
                continue

            key = _get_entity_key(entity)
            if key not in entities_map:
                entities_map[key] = (entity, [])
            if getattr(entity, "score", None) is not None:
                entities_map[key][1].append(entity.score)

        return [
            entity.model_copy(update={"score": _average_scores(scores)}) for entity, scores in entities_map.values()
        ]


class SingleEntityConsolidation(ConsolidationStrategy):
    """Consolidation strategy for a single entity."""

    def __init__(self, extractor: Callable[[Any], pydantic.BaseModel | None]):
        """Initialize SingleEntityConsolidation.

        :param extractor: Callable to extract a single entity from a raw chunk result.
        """
        self._extractor = extractor

    def consolidate(
        self,
        results: Sequence[Any],
        docs_offsets: list[tuple[int, int]],
    ) -> Sequence[pydantic.BaseModel | None]:
        """Consolidate single entities from chunks via majority vote.

        :param results: Raw chunk results.
        :param docs_offsets: Chunk offsets per document.
        :return: Winner entity per document.
        """
        consolidated_results: list[pydantic.BaseModel | None] = []
        for start, end in docs_offsets:
            chunk_results = [
                (self._extractor(res) if res is not None else None, i) for i, res in enumerate(results[start:end])
            ]
            consolidated_results.append(self._consolidate_single(chunk_results))
        return consolidated_results

    @staticmethod
    def _consolidate_single(entities_with_indices: list[tuple[_EntityType | None, int]]) -> _EntityType | None:
        """Majority vote for single entity with tie-breaking based on first occurrence."""
        if not entities_with_indices:
            return None

        counts: Counter[str] = Counter()
        key_to_entity: dict[str, _EntityType | None] = {}
        first_seen: dict[str, int] = {}
        scores_map: dict[str, list[float]] = defaultdict(list)

        for entity, i in entities_with_indices:
            key = _get_entity_key(entity) if entity is not None else "null"
            counts[key] += 1
            if key not in key_to_entity:
                key_to_entity[key] = entity
            if key not in first_seen:
                first_seen[key] = i
            if entity is not None and getattr(entity, "score", None) is not None:
                scores_map[key].append(entity.score)

        max_count = max(counts.values())
        winners = [k for k, count in counts.items() if count == max_count]
        winner_key = min(winners, key=lambda k: first_seen[k])

        winner_entity = key_to_entity[winner_key]
        if winner_entity is None:
            return None

        return winner_entity.model_copy(update={"score": _average_scores(scores_map[winner_key])})


class LabelScoreConsolidation(ConsolidationStrategy):
    """Consolidation strategy for classification tasks."""

    def __init__(
        self,
        labels: list[str],
        mode: Literal["single", "multi"],
        extractor: Callable[[Any], dict[str, float]],
    ):
        """Initialize LabelScoreConsolidation.

        :param labels: List of valid labels.
        :param mode: Classification mode ('single' or 'multi').
        :param extractor: Callable to extract label scores from a raw chunk result.
        """
        self.labels = labels
        self.mode = mode
        self.extractor = extractor

    def consolidate(
        self,
        results: Sequence[Any],
        docs_offsets: list[tuple[int, int]],
    ) -> Sequence[list[tuple[str, float]]]:
        """Consolidate label scores from chunks by averaging them.

        :param results: Raw chunk results.
        :param docs_offsets: Chunk offsets per document.
        :return: Sorted list of (label, score) pairs per document.
        """
        consolidated_results: list[list[tuple[str, float]]] = []
        for start, end in docs_offsets:
            chunk_results = [res for res in results[start:end] if res is not None]
            num_chunks = end - start
            label_scores = {label: 0.0 for label in self.labels}

            for res in chunk_results:
                scores = self.extractor(res)
                for label, score in scores.items():
                    if label in label_scores:
                        label_scores[label] += max(0.0, min(float(score), 1.0))

            avg_scores = sorted(
                ((label, score / num_chunks) for label, score in label_scores.items()),
                key=lambda x: x[1],
                reverse=True,
            )
            consolidated_results.append(avg_scores)

        return consolidated_results


class TextConsolidation(ConsolidationStrategy):
    """Consolidation strategy for tasks producing text (translation, summarization)."""

    def __init__(self, extractor: Callable[[Any], tuple[str, float | None]], joiner: str = "\n"):
        """Initialize TextConsolidation.

        :param extractor: Callable to extract (text, score) from a raw chunk result.
        :param joiner: String used to join text chunks.
        """
        self.extractor = extractor
        self.joiner = joiner

    def consolidate(
        self,
        results: Sequence[Any],
        docs_offsets: list[tuple[int, int]],
    ) -> Sequence[tuple[str, float | None]]:
        """Consolidate text chunks by joining them and averaging scores.

        :param results: Raw chunk results.
        :param docs_offsets: Chunk offsets per document.
        :return: Consolidated (text, average_score) per document.
        """
        consolidated_results: list[tuple[str, float | None]] = []
        for start, end in docs_offsets:
            texts: list[str] = []
            scores: list[float] = []

            for res in results[start:end]:
                if res is None:
                    continue
                text, score = self.extractor(res)
                texts.append(text)
                if score is not None:
                    scores.append(score)

            consolidated_results.append((self.joiner.join(texts).strip(), _average_scores(scores)))

        return consolidated_results


class QAConsolidation(ConsolidationStrategy):
    """Consolidation strategy for question answering."""

    def __init__(self, questions: list[str], extractor: Callable[[Any], Iterable[tuple[str, str, float | None]]]):
        """Initialize QAConsolidation.

        :param questions: List of questions.
        :param extractor: Callable to extract (question, answer, score) tuples from a raw chunk result.
        """
        self.questions = questions
        self.extractor = extractor

    def consolidate(
        self,
        results: Sequence[Any],
        docs_offsets: list[tuple[int, int]],
    ) -> Sequence[list[tuple[str, str, float | None]]]:
        """Consolidate QA pairs using sequential matching.

        Assumes that the i-th QA pair returned by the model corresponds to the i-th question in self.questions.

        :param results: Raw chunk results.
        :param docs_offsets: Chunk offsets per document.
        :return: Consolidated list of (question, answer, score) per document.
        """
        consolidated_results: list[list[tuple[str, str, float | None]]] = []
        for start, end in docs_offsets:
            qa_map: dict[str, tuple[list[str], list[float]]] = {q: ([], []) for q in self.questions}

            for res in results[start:end]:
                if res is None:
                    continue

                for i, (_, answer, score) in enumerate(self.extractor(res)):
                    if i < len(self.questions):
                        question = self.questions[i]
                        qa_map[question][0].append(answer)
                        if score is not None:
                            qa_map[question][1].append(score)

            consolidated_qa = [
                (q, " ".join(qa_map[q][0]).strip(), _average_scores(qa_map[q][1])) for q in self.questions
            ]
            consolidated_results.append(consolidated_qa)

        return consolidated_results


class MapScoreConsolidation(ConsolidationStrategy):
    """Consolidation strategy for map-based scores (e.g. sentiment analysis)."""

    def __init__(
        self,
        keys: Iterable[str],
        extractor: Callable[[Any], tuple[dict[str, float], float | None]],
    ):
        """Initialize MapScoreConsolidation.

        :param keys: Keys (aspects/labels) to average.
        :param extractor: Callable to extract (map_scores, overall_score) from a raw chunk result.
        """
        self.keys = list(keys)
        self.extractor = extractor

    def consolidate(
        self,
        results: Sequence[Any],
        docs_offsets: list[tuple[int, int]],
    ) -> Sequence[tuple[dict[str, float], float | None]]:
        """Consolidate map-based scores by averaging them.

        :param results: Raw chunk results.
        :param docs_offsets: Chunk offsets per document.
        :return: Consolidated (avg_map_scores, avg_overall_score) per document.
        """
        consolidated_results: list[tuple[dict[str, float], float | None]] = []
        for start, end in docs_offsets:
            num_chunks = end - start
            key_scores = {k: 0.0 for k in self.keys}
            overall_scores: list[float] = []

            for res in results[start:end]:
                if res is None:
                    continue
                m_scores, o_score = self.extractor(res)
                for k, s in m_scores.items():
                    if k in key_scores:
                        key_scores[k] += max(0.0, min(float(s), 1.0))
                if o_score is not None:
                    overall_scores.append(o_score)

            avg_map = {k: s / num_chunks for k, s in key_scores.items()}
            consolidated_results.append((avg_map, _average_scores(overall_scores)))

        return consolidated_results



"""Tests for Docling task."""
from pathlib import Path

import pytest
from docling.datamodel.base_models import InputFormat
from docling.datamodel.pipeline_options import PdfPipelineOptions, smolvlm_picture_description
from docling.document_converter import DocumentConverter, PdfFormatOption
from docling_core.types.doc.document import DescriptionAnnotation

from sieves import Doc, Pipeline
from sieves.serialization import Config
from sieves.tasks import Ingestion
from sieves.tasks.preprocessing.ingestion.docling_ import Docling


def test_run() -> None:
    resources = [Doc(uri=Path(__file__).parent.parent.parent.parent / "assets" / "1204.0162v2.pdf")]
    pipe = Pipeline(tasks=[Docling()])
    docs = list(pipe(resources))

    assert len(docs) == 1
    assert docs[0].text


def test_serialization() -> None:
    resources = [Doc(uri=Path(__file__).parent.parent.parent.parent / "assets" / "1204.0162v2.pdf")]
    pipe = Pipeline(tasks=[Docling()])
    docs = list(pipe(resources))

    config = pipe.serialize()
    version = Config.get_version()
    assert config.model_dump() == {
        "cls_name": "sieves.pipeline.core.Pipeline",
        "use_cache": {"is_placeholder": False, "value": True},
        "tasks": {
            "is_placeholder": False,
            "value": [
                {
                    "cls_name": "sieves.tasks.preprocessing.ingestion.docling_.Docling",
                    'batch_size': {'is_placeholder': False, "value": -1},
                    "converter": {"is_placeholder": True, "value": "docling.document_converter.DocumentConverter"},
                    "export_format": {"is_placeholder": False, "value": "markdown"},
                    "include_meta": {"is_placeholder": False, "value": False},
                    "task_id": {"is_placeholder": False, "value": "Docling"},
                    "version": version,
                    'condition': {'is_placeholder': False, 'value': None},
                }
            ],
        },
        "version": version,
    }

    deserialized_pipeline = Pipeline.deserialize(
        config=config, tasks_kwargs=[{"converter": None, "export_format": "markdown"}]
    )
    assert docs[0] == list(deserialized_pipeline(resources))[0]


@pytest.mark.slow
def test_image_description_with_local_vlm() -> None:
    """Test image description with local VLM model."""
    converter = DocumentConverter(
        format_options={
            InputFormat.PDF: PdfFormatOption(
                pipeline_options=PdfPipelineOptions(
                    do_picture_description=True,
                    picture_description_options=smolvlm_picture_description,
                    generate_picture_images=True,
                    images_scale=2.0
                )
            )
        }
    )

    resources = [Doc(uri=Path(__file__).parent.parent.parent.parent / "assets" / "1204.0162v2.pdf")]
    pipe = Pipeline(tasks=[Ingestion(converter=converter, include_meta=True)])
    docs = list(pipe(resources))

    assert len(docs) == 1
    assert docs[0].text
    pics_annotations = [pic.annotations for pic in docs[0].meta['Ingestion'].document.pictures]
    assert len(pics_annotations) == 6
    assert all(
        isinstance(pic_annotation, DescriptionAnnotation)
        for pic_annotations in pics_annotations for pic_annotation in pic_annotations
    )
    assert all(
        pic_annotation.text in docs[0].text
        for pic_annotations in pics_annotations for pic_annotation in pic_annotations
    )



# mypy: ignore-errors
from pathlib import Path
from tempfile import TemporaryDirectory

import datasets
import model2vec
import model2vec.inference
import model2vec.train
import numpy as np
import pytest
import setfit

from sieves import Doc, Pipeline
from sieves.model_wrappers import ModelType
from sieves.serialization import Config
from sieves.tasks import DistillationFramework
from sieves.tasks.predictive import classification


def _get_docs() -> list[Doc]:
    science_text = (
        "Scientists report that plasma is a state of matter. They published an academic paper. This is about science -"
        " scientists, papers, experiments, laws of nature."
    )
    politics_text = (
        "A new law has been passed. The opposition doesn't support it, but parliament has voted on it. This is about "
        "politics - parliament, laws, parties, politicians."
    )

    return [
        *[Doc(text=f"{i}. {science_text}") for i in range(5)],
        *[Doc(text=f"{i}. {politics_text}") for i in range(5)],
    ]


@pytest.mark.parametrize("batch_runtime", (ModelType.huggingface,), indirect=["batch_runtime"])
@pytest.mark.parametrize("distillation_framework", DistillationFramework.all())
def test_distillation_classification(batch_runtime, distillation_framework) -> None:
    seed = 42
    base_model_id = "sentence-transformers/paraphrase-mpnet-base-v2"
    if distillation_framework == DistillationFramework.model2vec:
        base_model_id = "minishlab/potion-base-32M"

    docs = _get_docs()

    with TemporaryDirectory() as tmp_dir:
        classifier = classification.Classification(
            task_id="classifier",
            model=batch_runtime.model,
            model_settings=batch_runtime.model_settings,
            batch_size=batch_runtime.batch_size,
            labels={
                "science": "Topics related to scientific disciplines and research",
                "politics": "Topics related to government, elections, and political systems",
            },
        )
        pipe = Pipeline([classifier])
        docs = list(pipe(docs))

        if distillation_framework == DistillationFramework.sentence_transformers:
            with pytest.raises(NotImplementedError):
                classifier.distill(
                    base_model_id=base_model_id,
                    framework=distillation_framework,
                    output_path=Path(tmp_dir),
                    val_frac=0.5,
                    seed=seed,
                    data=docs
                )
        else:
            classifier.distill(
                base_model_id=base_model_id,
                framework=distillation_framework,
                output_path=Path(tmp_dir),
                val_frac=0.5,
                seed=seed,
                data=docs
            )


        # Ensure equality of saved with original dataset.
        hf_dataset = classifier.to_hf_dataset(docs)
        hf_dataset = classifier._split_dataset(hf_dataset, 0.5, 0.5, seed)
        hf_dataset_loaded = datasets.DatasetDict.load_from_disk(Path(tmp_dir) / "data")
        for split in ("train", "val"):
            assert hf_dataset_loaded[split].info == hf_dataset[split].info
            assert hf_dataset_loaded[split]["text"] == hf_dataset[split]["text"]
            assert hf_dataset_loaded[split]["labels"] == hf_dataset[split]["labels"]

        # Assert predictions of distilled models look as expected.
        test_sents = ["This is about the galaxy and laws of nature.", "This is about political election and lobbying."]

        match distillation_framework:
            case DistillationFramework.setfit:
                model = setfit.SetFitModel.from_pretrained(tmp_dir)
                preds = model.predict(test_sents, as_numpy=True)
                assert preds.shape == (2, 2)

            case DistillationFramework.model2vec:
                model = model2vec.inference.StaticModelPipeline.from_pretrained(tmp_dir)
                preds = model.predict(test_sents)
                assert set(np.unique(preds).tolist()) <= {"science", "politics"}
                assert preds.shape[0] == 2
                assert preds.shape[1] in (1, 2)


@pytest.mark.parametrize("batch_runtime", [ModelType.huggingface], indirect=["batch_runtime"])
def test_serialization(batch_runtime) -> None:
    seed = 42
    dir_path: str | None
    docs = _get_docs()

    with TemporaryDirectory() as tmp_dir:
        classifier = classification.Classification(
            task_id="classifier",
            model=batch_runtime.model,
            model_settings=batch_runtime.model_settings,
            batch_size=batch_runtime.batch_size,
            labels={
                "science": "Topics related to scientific disciplines and research",
                "politics": "Topics related to government, elections, and political systems",
            },
        )
        pipe = Pipeline(classifier)
        docs = list(pipe(docs))

        classifier.distill(
            base_model_id="sentence-transformers/paraphrase-mpnet-base-v2",
            framework=DistillationFramework.setfit,
            output_path=Path(tmp_dir),
            val_frac=.5,
            seed=seed,
            data=docs
        )

    config = pipe.serialize()
    assert config.model_dump() == {'cls_name': 'sieves.pipeline.core.Pipeline',
 'tasks': {'is_placeholder': False,
           'value': [{'cls_name': 'sieves.tasks.predictive.classification.core.Classification',
                      'fewshot_examples': {'is_placeholder': False,
                                           'value': ()},
                      'batch_size': {'is_placeholder': False, 'value': -1},
                      'condition': {'is_placeholder': False, 'value': None},
                      'model_settings': {'is_placeholder': False,
                                              'value': {
                                                        'config_kwargs': None,
                                                        'inference_kwargs': None,
                                                        'init_kwargs': None,
                                                        'strict': True,
                                                        'inference_mode': None,}},
                      'include_meta': {'is_placeholder': False, 'value': True},
                      'labels': {'is_placeholder': False,
                                             'value': {'politics': 'Topics '
                                                                   'related to '
                                                                   'government, '
                                                                   'elections, '
                                                                   'and '
                                                                   'political '
                                                                   'systems',
                                                       'science': 'Topics '
                                                                  'related to '
                                                                  'scientific '
                                                                  'disciplines '
                                                                  'and '
                                                                  'research'}},
                      'model': {'is_placeholder': True,
                                'value': 'transformers.pipelines.zero_shot_classification.ZeroShotClassificationPipeline'},
                      'prompt_instructions': {'is_placeholder': False,
                                          'value': None},
                      'task_id': {'is_placeholder': False,
                                  'value': 'classifier'},
                      'version': Config.get_version()},
                     ]},
 'use_cache': {'is_placeholder': False, 'value': True},
 'version': Config.get_version()}

    Pipeline.deserialize(config=config, tasks_kwargs=[{"model": batch_runtime.model}])



"""Core evaluation logic and base classes."""

from __future__ import annotations

import dataclasses
from typing import Any

from sieves.data import Doc


@dataclasses.dataclass
class TaskEvaluationReport[T]:
    """Report containing evaluation results."""

    metrics: dict[str, float]
    task_id: str
    failures: list[Doc] = dataclasses.field(default_factory=list)

    def summary(self) -> str:
        """Return a string summary of the report.

        :return: Summary string.
        """
        metrics_str = ", ".join(f"{k}: {v:.4f}" for k, v in self.metrics.items())
        return f"Task: {self.task_id} | Metrics: {metrics_str} | Failures: {len(self.failures)}"


@dataclasses.dataclass
class PipelineEvaluationReport:
    """Report containing evaluation results for a pipeline."""

    reports: dict[str, TaskEvaluationReport[Any]]

    def summary(self) -> str:
        """Return a string summary of the pipeline report.

        :return: Summary string.
        """
        summaries = [report.summary() for report in self.reports.values()]
        return "\n".join(summaries)

    def __getitem__(self, task_id: str) -> TaskEvaluationReport[Any]:
        """Get report for specified task.

        :param task_id: Task ID.
        :return: Evaluation report.
        """
        return self.reports[task_id]



# mypy: ignore-errors
import pytest

from sieves import Pipeline, Doc
from sieves.model_wrappers import ModelType, ModelSettings
from sieves.serialization import Config
from sieves.tasks.predictive import relation_extraction
from sieves.tasks.predictive.schemas.relation_extraction import RelationEntity, RelationTriplet, Result
from sieves.tests.conftest import _make_runtime


@pytest.mark.parametrize(
    "batch_runtime",
    relation_extraction.RelationExtraction.supports(),
    indirect=["batch_runtime"],
)
@pytest.mark.parametrize("fewshot", [True, False])
def test_run(relation_extraction_docs, batch_runtime, fewshot) -> None:
    # --8<-- [start:re-usage]
    relations = {
        "works_for": "A person works for a company or organization.",
        "located_in": "A place or organization is located in a city, country, or region.",
        "founded": "A person founded a company or organization.",
    }

    fewshot_examples = [
        relation_extraction.FewshotExample(
            text="Henri Dunant founded the Red Cross in Geneva.",
            triplets=[
                RelationTriplet(
                    head=RelationEntity(text="Henri Dunant", entity_type="PERSON"),
                    relation="founded",
                    tail=RelationEntity(text="Red Cross", entity_type="ORGANIZATION"),
                    score=1.0,
                ),
                RelationTriplet(
                    head=RelationEntity(text="Red Cross", entity_type="ORGANIZATION"),
                    relation="located_in",
                    tail=RelationEntity(text="Geneva", entity_type="LOCATION"),
                    score=1.0,
                ),
            ],
        ),
        relation_extraction.FewshotExample(
            text="Eglantyne Jebb founded Save the Children in London.",
            triplets=[
                RelationTriplet(
                    head=RelationEntity(text="Eglantyne Jebb", entity_type="PERSON"),
                    relation="founded",
                    tail=RelationEntity(text="Save the Children", entity_type="ORGANIZATION"),
                    score=1.0,
                ),
                RelationTriplet(
                    head=RelationEntity(text="Save the Children", entity_type="ORGANIZATION"),
                    relation="located_in",
                    tail=RelationEntity(text="London", entity_type="LOCATION"),
                    score=1.0,
                ),
            ],
        ),
    ]

    fewshot_args = {"fewshot_examples": fewshot_examples} if fewshot else {}

    task = relation_extraction.RelationExtraction(
        relations=relations,
        model=batch_runtime.model,
        model_settings=batch_runtime.model_settings,
        batch_size=batch_runtime.batch_size,
        entity_types=["PERSON", "ORGANIZATION", "LOCATION"],
        **fewshot_args
    )

    pipe = Pipeline(task)
    docs = list(pipe(relation_extraction_docs))
    # --8<-- [end:re-usage]

    assert len(docs) == 2
    for doc in docs:
        assert "RelationExtraction" in doc.results
        res = doc.results["RelationExtraction"]
        assert isinstance(res, relation_extraction.Result)
        assert len(res.triplets)

        assert "RelationExtraction" in doc.meta
        assert "raw" in doc.meta["RelationExtraction"]
        assert "usage" in doc.meta["RelationExtraction"]
        assert "usage" in doc.meta

        print(f"Output: {doc.results['RelationExtraction']}")
        print(f"Raw output: {doc.meta['RelationExtraction']['raw']}")
        print(f"Usage: {doc.meta['RelationExtraction']['usage']}")
        print(f"Total Usage: {doc.meta['usage']}")

        # Verify schema.
        for triplet in res.triplets:
            assert isinstance(triplet, relation_extraction.RelationTriplet)
            assert isinstance(triplet.head, relation_extraction.RelationEntity)
            assert isinstance(triplet.tail, relation_extraction.RelationEntity)
            assert isinstance(triplet.relation, str)
            assert triplet.relation.lower() in ["works_for", "located_in", "founded"]

            # Verify content.
            assert triplet.head.text in doc.text
            assert triplet.tail.text in doc.text

    with pytest.raises(NotImplementedError):
        pipe["RelationExtraction"].distill(None, None, None, None, None, None, None, None)

    if fewshot:
        _to_hf_dataset(task, docs)


@pytest.mark.parametrize("batch_runtime", [ModelType.dspy], indirect=["batch_runtime"])
def test_serialization(relation_extraction_docs, batch_runtime) -> None:
    pipe = Pipeline(
        relation_extraction.RelationExtraction(
            relations=["works_for", "located_in", "founded"],
            model=batch_runtime.model,
            model_settings=batch_runtime.model_settings,
            batch_size=batch_runtime.batch_size,
            entity_types=["PERSON", "ORGANIZATION", "LOCATION"]
        )
    )

    config = pipe.serialize()
    assert config.model_dump() == {
        'cls_name': 'sieves.pipeline.core.Pipeline',
        'tasks': {
            'is_placeholder': False,
            'value': [
                {
                    'cls_name': 'sieves.tasks.predictive.relation_extraction.core.RelationExtraction',
                    'relations': {'is_placeholder': False, 'value': ['works_for', 'located_in', 'founded']},
                    'entity_types': {'is_placeholder': False, 'value': ['PERSON', 'ORGANIZATION', 'LOCATION']},
                    'fewshot_examples': {'is_placeholder': False, 'value': ()},
                    'batch_size': {'is_placeholder': False, 'value': -1},
                    'model_settings': {
                        'is_placeholder': False,
                        'value': {
                            'config_kwargs': None,
                            'inference_kwargs': None,
                            'init_kwargs': None,
                            'strict': True,
                            'inference_mode': None,
                        }
                    },
                    'include_meta': {'is_placeholder': False, 'value': True},
                    'model': {'is_placeholder': True, 'value': 'dspy.clients.lm.LM'},
                    'prompt_instructions': {'is_placeholder': False,
                                          'value': None},
                    'task_id': {'is_placeholder': False, 'value': 'RelationExtraction'},
                    'condition': {'is_placeholder': False, 'value': None},
                    'version': Config.get_version()
                }
            ]
        },
        'use_cache': {'is_placeholder': False, 'value': True},
        'version': Config.get_version()
    }

    Pipeline.deserialize(
        config=config,
        tasks_kwargs=[{"model": batch_runtime.model}],
    )


def _to_hf_dataset(task: relation_extraction.RelationExtraction, docs: list[Doc]) -> None:
    """Tests whether conversion to HF dataset works as expected.

    :param task: RelationExtraction task instance.
    :param docs: List of documents to convert.
    """
    dataset = task.to_hf_dataset(docs)
    assert all([key in dataset.features for key in ("text", "triplets")])
    assert len(dataset) == 2
    dataset_records = list(dataset)
    for rec in dataset_records:
        # Hugging Face datasets stores sequence of dicts as dict of lists (column-oriented)
        triplets = rec["triplets"]
        assert isinstance(triplets, dict)
        assert "head" in triplets
        assert "relation" in triplets
        assert "tail" in triplets

        # Verify consistent lengths
        assert len(triplets["head"]) == len(triplets["relation"]) == len(triplets["tail"])

        if len(triplets["head"]) > 0:
            # Check structure of the first item
            head = triplets["head"][0]
            assert isinstance(head, dict)
            assert "text" in head
            assert "entity_type" in head
            assert isinstance(triplets["relation"][0], str)

        assert isinstance(rec["text"], str)

    with pytest.raises(KeyError):
        task.to_hf_dataset([Doc(text="This is a dummy text.")])


@pytest.mark.parametrize(
    "batch_runtime",
    [ModelType.dspy, ModelType.langchain, ModelType.outlines, ModelType.gliner],
    indirect=["batch_runtime"],
)
def test_inference_mode_override(batch_runtime) -> None:
    """Test that inference_mode parameter overrides the default value."""
    dummy = "dummy_inference_mode"

    task = relation_extraction.RelationExtraction(
        relations=["works_for"],
        model=batch_runtime.model,
        model_settings=ModelSettings(inference_mode=dummy),
        batch_size=batch_runtime.batch_size,
    )

    assert task._bridge.inference_mode == dummy


@pytest.mark.parametrize(
    "batch_runtime",
    relation_extraction.RelationExtraction.supports(),
    indirect=["batch_runtime"],
)
def test_run_with_list_relations(relation_extraction_docs, batch_runtime) -> None:
    """Test RelationExtraction with list format relations."""
    relations_list = ["works_for", "located_in"]

    pipe = Pipeline(
        relation_extraction.RelationExtraction(
            relations=relations_list,
            model=batch_runtime.model,
            model_settings=batch_runtime.model_settings,
            batch_size=batch_runtime.batch_size,
            entity_types=["PERSON", "ORGANIZATION", "LOCATION"]
        )
    )
    docs = list(pipe(relation_extraction_docs))

    assert len(docs) == 2
    for doc in docs:
        assert "RelationExtraction" in doc.results


def test_gliner_warning() -> None:
    """Test that GliNER2 issues a warning when entity_types are provided."""
    batch_runtime_gliner = _make_runtime(model_type=ModelType.gliner, batch_size=-1)
    relations = ["works_for"]

    with pytest.warns(UserWarning, match="GliNER2 backend does not support entity type constraints"):
        relation_extraction.RelationExtraction(
            relations=relations,
            model=batch_runtime_gliner.model,
            entity_types=["PERSON"],
            task_id="re",
        )


@pytest.mark.parametrize("batch_runtime", [ModelType.gliner], indirect=["batch_runtime"])
def test_evaluation(batch_runtime) -> None:
    """Test evaluation for relation extraction without running pipeline."""
    task = relation_extraction.RelationExtraction(
        relations=["founded", "located_in"], model=batch_runtime.model, task_id="re"
    )

    # 1. Full overlap
    doc_full = Doc(text="Henri founded Red Cross in Geneva.")
    res_full = Result(triplets=[
        RelationTriplet(
            head=RelationEntity(text="Henri", entity_type="PERSON"),
            relation="founded",
            tail=RelationEntity(text="Red Cross", entity_type="ORGANIZATION")
        ),
        RelationTriplet(
            head=RelationEntity(text="Red Cross", entity_type="ORGANIZATION"),
            relation="located_in",
            tail=RelationEntity(text="Geneva", entity_type="LOCATION")
        )
    ])
    doc_full.results["re"] = res_full
    doc_full.gold["re"] = res_full
    report_full = task.evaluate([doc_full])
    assert report_full.metrics[task.metric] == 1.0

    # 2. Partial overlap
    doc_partial = Doc(text="Henri founded Red Cross in Geneva.")
    # Pred has both
    res_pred = res_full
    # Gold has only one
    res_gold = Result(triplets=[
        RelationTriplet(
            head=RelationEntity(text="Henri", entity_type="PERSON"),
            relation="founded",
            tail=RelationEntity(text="Red Cross", entity_type="ORGANIZATION")
        )
    ])
    doc_partial.results["re"] = res_pred
    doc_partial.gold["re"] = res_gold
    report_partial = task.evaluate([doc_partial])
    # TP=1 (founded), FP=1 (located_in), FN=0 -> Precision=0.5, Recall=1.0 -> F1=0.666...
    assert 0.6 < report_partial.metrics[task.metric] < 0.7

    # 3. No overlap
    doc_none = Doc(text="Henri founded Red Cross.")
    res_none_pred = Result(triplets=[
        RelationTriplet(
            head=RelationEntity(text="Henri", entity_type="PERSON"),
            relation="founded",
            tail=RelationEntity(text="Red Cross", entity_type="ORGANIZATION")
        )
    ])
    res_none_gold = Result(triplets=[
        RelationTriplet(
            head=RelationEntity(text="Eglantyne", entity_type="PERSON"),
            relation="founded",
            tail=RelationEntity(text="Save the Children", entity_type="ORGANIZATION")
        )
    ])
    doc_none.results["re"] = res_none_pred
    doc_none.gold["re"] = res_none_gold
    report_none = task.evaluate([doc_none])
    assert report_none.metrics[task.metric] == 0.0



"""Bridge for GLiNER2 models."""

from __future__ import annotations

import types
import typing
from collections import defaultdict
from collections.abc import Sequence
from typing import Any, Literal, override

import gliner2
import pydantic

from sieves.data import Doc
from sieves.model_wrappers import ModelType, gliner_
from sieves.model_wrappers.types import ModelSettings
from sieves.tasks.predictive.bridges import Bridge
from sieves.tasks.predictive.schemas.classification import ResultMultiLabel, ResultSingleLabel
from sieves.tasks.predictive.schemas.information_extraction import (
    ResultMulti as IEResultMulti,
)
from sieves.tasks.predictive.schemas.information_extraction import (
    ResultSingle as IEResultSingle,
)
from sieves.tasks.predictive.schemas.ner import Entity as NEREntity
from sieves.tasks.predictive.schemas.ner import Result as NERResult
from sieves.tasks.predictive.schemas.relation_extraction import (
    RelationEntity as RERelationEntity,
)
from sieves.tasks.predictive.schemas.relation_extraction import (
    RelationTriplet as RERelationTriplet,
)
from sieves.tasks.predictive.schemas.relation_extraction import (
    Result as RERelationResult,
)
from sieves.tasks.predictive.utils import convert_to_signature


class GliNERBridge(Bridge[gliner2.inference.engine.Schema, gliner_.Result, gliner_.InferenceMode]):
    """Bridge for GLiNER2 models."""

    def __init__(
        self,
        task_id: str,
        prompt_instructions: str | None,
        prompt_signature: type[pydantic.BaseModel],
        model_settings: ModelSettings,
        inference_mode: gliner_.InferenceMode,
        mode: Literal["multi", "single"] = "multi",
    ):
        """Initialize GLiNER2 bridge.

        Important: currently only GLiNER2 schemas/structures with one key each are supported. We do NOT support
        composite requests like `create_schema().entities().classification(). ...`.

        :param task_id: Task ID.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param prompt_signature: Unified Pydantic prompt signature.
        :param model_settings: Model settings including inference_mode.
        :param mode: Extraction mode. If "multi", all occurrences of the entity are extracted. If "single", exactly one
            (or no) entity is extracted. Only used if `inference_mode == InferenceMode.structure`, ignored otherwise.
        """
        super().__init__(
            task_id=task_id,
            prompt_instructions=prompt_instructions,
            overwrite=False,
            model_settings=model_settings,
            prompt_signature=prompt_signature,
            model_type=ModelType.gliner,
        )
        self._inference_mode = inference_mode
        self._mode = mode

    @override
    @property
    def model_type(self) -> ModelType:
        return ModelType.gliner

    @override
    @property
    def prompt_signature(self) -> gliner2.inference.engine.Schema | gliner2.inference.engine.StructureBuilder:
        # Map internal inference mode to GliNER utility mode.
        mode_map = {
            gliner_.InferenceMode.classification: "classification",
            gliner_.InferenceMode.entities: "entities",
            gliner_.InferenceMode.structure: "structure",
            gliner_.InferenceMode.relations: "relations",
        }
        mode = mode_map[self.inference_mode]

        model_cls = self._pydantic_signature
        # Unwrap unified signature container if necessary.
        if mode == "structure" or mode == "entities":
            if "entities" in model_cls.model_fields:
                model_cls = model_cls.model_fields["entities"].annotation
                if typing.get_origin(model_cls) is list:
                    model_cls = typing.get_args(model_cls)[0]
            elif "entity" in model_cls.model_fields:
                model_cls = model_cls.model_fields["entity"].annotation
                # Handle Optional/Union.
                if typing.get_origin(model_cls) in (typing.Union, types.UnionType):
                    args = typing.get_args(model_cls)
                    model_cls = [t for t in args if t is not type(None)][0]

        elif mode == "relations":
            if "triplets" in model_cls.model_fields:
                model_cls = model_cls.model_fields["triplets"].annotation
                if typing.get_origin(model_cls) is list:
                    model_cls = typing.get_args(model_cls)[0]

        kwargs = {"mode": mode}
        if mode == "classification":
            kwargs["task"] = self._task_id

        prompt_signature = convert_to_signature(
            model_cls=model_cls,
            model_type=ModelType.gliner,
            **kwargs,
        )
        assert isinstance(prompt_signature, gliner2.inference.engine.Schema | gliner2.inference.engine.StructureBuilder)

        return prompt_signature

    @property
    def prompt_signature_pydantic(self) -> type[pydantic.BaseModel] | None:
        """Return Pydantic model representation of GLiNER2 schema.

        :return: Pydantic model representation of GLiNER2 schema.
        """
        return self._pydantic_signature

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        # GLiNER2 doesn't support custom instructions.
        return ""

    @override
    @property
    def inference_mode(self) -> gliner_.InferenceMode:
        return self._model_settings.inference_mode or self._inference_mode

    @override
    def integrate(self, results: Sequence[gliner_.Result], docs: list[Doc]) -> list[Doc]:
        for doc, result in zip(docs, results):
            match self._inference_mode:
                # Used by: Classification
                case gliner_.InferenceMode.classification:
                    if self._mode == "multi":
                        label_scores: list[tuple[str, float]] = []
                        for res in sorted(result, key=lambda x: x["score"], reverse=True):
                            assert isinstance(res, dict)
                            label_scores.append((res["label"], res["score"]))
                        doc.results[self._task_id] = ResultMultiLabel(label_scores=label_scores)

                    else:
                        doc.results[self._task_id] = ResultSingleLabel(
                            label=result[0]["label"], score=result[0]["score"]
                        )

                # Used by: NER
                case gliner_.InferenceMode.entities:
                    if isinstance(result, list):
                        # Result is a list of entity dictionaries.
                        doc.results[self._task_id] = NERResult(
                            entities=[
                                NEREntity.model_validate(
                                    {k if k != "confidence" else "score": v for k, v in res.items()}
                                )
                                for res in result
                            ],
                            text=doc.text or "",
                        )
                    else:
                        doc.results[self._task_id] = result

                # Used by: InformationExtraction
                case gliner_.InferenceMode.structure:
                    if not result:
                        if self._mode == "multi":
                            doc.results[self._task_id] = IEResultMulti(entities=[])
                        else:
                            doc.results[self._task_id] = IEResultSingle(entity=None)
                        continue

                    entity_type_name = list(result.keys())[0]
                    # Unified signature container unwrapping
                    validation_model = self.prompt_signature_pydantic
                    if validation_model and "entities" in validation_model.model_fields:
                        # Multi-mode container
                        validation_model = validation_model.model_fields["entities"].annotation
                        if hasattr(validation_model, "__args__"):
                            validation_model = validation_model.__args__[0]

                    elif validation_model and "entity" in validation_model.model_fields:
                        # Single-mode container
                        validation_model = validation_model.model_fields["entity"].annotation
                        if hasattr(validation_model, "__args__"):
                            # Filter out None
                            validation_model = [t for t in validation_model.__args__ if t is not type(None)][0]

                    assert validation_model is not None and issubclass(validation_model, pydantic.BaseModel)

                    extracted_entities: list[pydantic.BaseModel] = []
                    for entity in result[entity_type_name]:
                        # Map fields and include score.
                        entity_data = {key: value["text"] for key, value in entity.items()}
                        # Calculate average confidence across all fields.
                        confidences = [v["confidence"] for v in entity.values() if "confidence" in v]
                        entity_data["score"] = sum(confidences) / len(confidences) if confidences else None

                        extracted_entities.append(validation_model.model_validate(entity_data))

                    # This covers single/multi mode for information extraction.
                    if self._mode == "multi":
                        doc.results[self._task_id] = IEResultMulti(entities=extracted_entities)
                    else:
                        doc.results[self._task_id] = IEResultSingle(
                            entity=extracted_entities[0] if extracted_entities else None
                        )

                # Used by: RelationExtraction
                case gliner_.InferenceMode.relations:
                    triplets: list[RERelationTriplet] = []
                    # GliNER2 relations output is a dict mapping task ids to relation types to lists of triplets:
                    # {'task_id': {'relation_type': [{'head': {...}, 'tail': {...}}, ...]}}  # noqa: ERA001
                    relation_data = result.get(self._task_id) or result.get("relation_extraction", {})
                    assert isinstance(relation_data, dict)

                    for rel_type, triplets_list in relation_data.items():
                        for triplet_data in triplets_list:
                            # Calculate score as average of head and tail confidence if they exist.
                            head_conf = triplet_data.get("head", {}).get("confidence")
                            tail_conf = triplet_data.get("tail", {}).get("confidence")
                            confidences = [c for c in (head_conf, tail_conf) if c is not None]
                            score = sum(confidences) / len(confidences) if confidences else None

                            triplets.append(
                                RERelationTriplet(
                                    head=RERelationEntity(
                                        text=triplet_data["head"]["text"],
                                        entity_type="UNKNOWN",
                                    ),
                                    relation=rel_type,
                                    tail=RERelationEntity(
                                        text=triplet_data["tail"]["text"],
                                        entity_type="UNKNOWN",
                                    ),
                                    score=score,
                                )
                            )

                    doc.results[self._task_id] = RERelationResult(triplets=triplets)

        return docs

    @override
    def consolidate(
        self, results: Sequence[gliner_.Result], docs_offsets: list[tuple[int, int]]
    ) -> Sequence[gliner_.Result]:
        consolidated_results: list[gliner_.Result] = []

        # Determine label scores for chunks per document.
        for doc_offset in docs_offsets:
            scores: dict[Any, float] = defaultdict(lambda: 0.0)
            entities: dict[str, list[str] | dict[str, str | list[str]]] = {}
            all_entities_list: list[dict[str, Any]] = []

            # In single mode for structure, we want to keep track of the highest confidence entity.
            highest_confidence_entity: dict[str, Any] | None = None
            max_confidence: float = -1.0

            for res in results[doc_offset[0] : doc_offset[1]]:
                match self._inference_mode:
                    case gliner_.InferenceMode.classification:
                        keys = list(res.keys())
                        assert len(keys) == 1, "Composite GliNER2 schemas are not supported."
                        extracted_res = res[keys[0]]

                        # In case of single-label: pad to list so that we can process in a unified way.
                        if isinstance(extracted_res, dict):
                            extracted_res = [extracted_res]

                        for entry in extracted_res:
                            # GliNER might use two different structures here, depending on the version.
                            if "label" in entry:
                                scores[entry["label"]] += entry["confidence"]
                            else:
                                assert isinstance(entry, dict)
                                keys = list(entry.keys())
                                assert len(keys) == 1, "Composite GliNER2 schemas are not supported."
                                for label, confidence in entry[keys[0]]:
                                    scores[label] += confidence

                    case gliner_.InferenceMode.entities:
                        for entity_type in res["entities"]:
                            items = res["entities"][entity_type]
                            if items:
                                if isinstance(items[0], dict):
                                    # Flatten into a list of entities (dicts).
                                    for item in items:
                                        all_entities_list.append({"entity_type": entity_type, **item})
                                else:
                                    if entity_type not in entities:
                                        entities[entity_type] = []
                                    relevant_entities: list[str] = entities[entity_type]  # type: ignore[assignment]
                                    relevant_entities.extend(items)

                    case gliner_.InferenceMode.structure:
                        for entity_type in res:
                            if entity_type not in entities:
                                entities[entity_type] = []
                            relevant_entities_struct: list[Any] = entities[entity_type]  # type: ignore[assignment]

                            # Add entities from this chunk.
                            for entity in res[entity_type]:
                                # If single mode: we already have logic to track max_confidence below.
                                # But we also need to store all entities for general structure if not single?
                                # Actually, GLiNER structure mode output is a bit different.
                                relevant_entities_struct.append(entity)

                            if self._mode == "single":
                                for entity in res[entity_type]:
                                    # Calculate average confidence for the entity structure.
                                    assert isinstance(entity, dict)
                                    confidences = [
                                        field_val["confidence"]
                                        for field_val in entity.values()
                                        if "confidence" in field_val
                                    ]
                                    if confidences:
                                        avg_conf = sum(confidences) / len(confidences)
                                        if avg_conf > max_confidence:
                                            max_confidence = avg_conf
                                            # We need to preserve the confidence in the structure
                                            # so that integrate can pick it up.
                                            highest_confidence_entity = {entity_type: [entity]}

                    case gliner_.InferenceMode.relations:
                        # Collect all triplets from all chunks for this document.
                        # Simple unification for now.
                        relation_data = res.get(self._task_id) or res.get("relation_extraction", {})
                        assert isinstance(relation_data, dict)

                        for rel_type, triplets_list in relation_data.items():
                            for triplet in triplets_list:
                                key = (
                                    triplet["head"]["text"],
                                    rel_type,
                                    triplet["tail"]["text"],
                                    triplet["head"]["start"],
                                    triplet["tail"]["start"],
                                )
                                if key not in scores:  # Use scores dict as a seen set for convenience.
                                    scores[key] = 1.0
                                    all_entities_list.append({"relation": rel_type, **triplet})

            match self._inference_mode:
                case gliner_.InferenceMode.classification:
                    # Ensure that all labels have been assigned - GLiNER2 is sometimes negligent about this.
                    # We extract labels from our Pydantic signature fields (excluding 'score' if present).
                    labels = [name for name in self._pydantic_signature.model_fields if name != "score"]
                    for label in labels:
                        if label not in scores:
                            scores[label] = 0.0

                    # Average score, sort in descending order.
                    sorted_scores: list[dict[str, str | float]] = sorted(
                        (
                            {"label": attr, "score": score / (doc_offset[1] - doc_offset[0])}
                            for attr, score in scores.items()
                        ),
                        key=lambda x: x["score"],
                        reverse=True,
                    )

                    consolidated_results.append(sorted_scores)

                case gliner_.InferenceMode.entities | gliner_.InferenceMode.structure | gliner_.InferenceMode.relations:
                    if self._inference_mode == gliner_.InferenceMode.structure and self._mode == "single":
                        if highest_confidence_entity:
                            consolidated_results.append(highest_confidence_entity)

                        else:
                            # Use the entity name from the pydantic signature (unwrapped).
                            model_cls = self._pydantic_signature

                            if "entity" in model_cls.model_fields:
                                model_cls = model_cls.model_fields["entity"].annotation
                                if typing.get_origin(model_cls) in (
                                    typing.Union,
                                    types.UnionType,
                                ):
                                    model_cls = [t for t in typing.get_args(model_cls) if t is not type(None)][0]

                            consolidated_results.append({model_cls.__name__: []})

                    elif all_entities_list:
                        if self._inference_mode == gliner_.InferenceMode.relations:
                            # Reconstruct the dict structure for integrate.
                            reconstructed: dict[str, list[Any]] = defaultdict(list)
                            for item in all_entities_list:
                                rel_type = item.pop("relation")
                                reconstructed[rel_type].append(item)
                            # Ensure we use the current task_id as the key so that integrate() can find it.
                            consolidated_results.append({self._task_id: dict(reconstructed)})
                        else:
                            consolidated_results.append(all_entities_list)
                    else:
                        consolidated_results.append(entities)

        return consolidated_results



"""Distillation framework types."""

from __future__ import annotations

import enum
from typing import Literal


class DistillationFramework(enum.Enum):
    """Available distillation frameworks."""

    model2vec = "model2vec"
    sentence_transformers = "sentence_transformers"
    setfit = "setfit"

    @classmethod
    def all(cls) -> tuple[DistillationFramework, ...]:
        """Return all available model types.

        :return tuple[ModelType, ...]: All available model types.
        """
        return tuple(dist_type for dist_type in DistillationFramework)


DistillationFrameworkLiteral = Literal[*DistillationFramework.all()]  # type: ignore[valid-type]



"""Tests for predictive utilities and consolidation strategies."""

import pydantic
import pytest

from sieves.tasks.predictive.consolidation import (
    MultiEntityConsolidation,
    SingleEntityConsolidation,
)


class DummyEntity(pydantic.BaseModel):
    """Dummy entity for testing."""

    name: str
    label: str
    score: float | None = None


class NonFrozenEntity(pydantic.BaseModel):
    """Non-frozen entity for testing."""

    name: str
    score: float | None = None


class NestedEntity(pydantic.BaseModel):
    """Nested entity for testing."""

    name: str
    details: DummyEntity
    score: float | None = None


def test_multi_entity_consolidation_empty():
    strategy = MultiEntityConsolidation(extractor=lambda x: x)
    # 0 chunks for 1 document
    assert strategy.consolidate([], [(0, 0)]) == [[]]


def test_multi_entity_consolidation_single():
    entity = DummyEntity(name="test", label="LABEL", score=0.8)
    strategy = MultiEntityConsolidation(extractor=lambda x: [x])
    results = strategy.consolidate([entity], [(0, 1)])
    assert len(results) == 1
    assert len(results[0]) == 1
    assert results[0][0].name == "test"
    assert results[0][0].score == 0.8


def test_multi_entity_consolidation_deduplication_and_averaging():
    entities = [
        DummyEntity(name="test", label="LABEL", score=0.8),
        DummyEntity(name="test", label="LABEL", score=0.6),
        DummyEntity(name="other", label="LABEL", score=0.9),
    ]
    strategy = MultiEntityConsolidation(extractor=lambda x: [x])
    results = strategy.consolidate(entities, [(0, 3)])
    assert len(results) == 1
    doc_results = results[0]
    assert len(doc_results) == 2

    # Sort results by name for deterministic testing
    doc_results = sorted(doc_results, key=lambda x: x.name)

    assert doc_results[0].name == "other"
    assert doc_results[0].score == pytest.approx(0.9)

    assert doc_results[1].name == "test"
    assert doc_results[1].score == pytest.approx(0.7)


def test_multi_entity_consolidation_mixed_scores():
    entities = [
        DummyEntity(name="test", label="LABEL", score=0.8),
        DummyEntity(name="test", label="LABEL", score=None),
    ]
    strategy = MultiEntityConsolidation(extractor=lambda x: [x])
    results = strategy.consolidate(entities, [(0, 2)])
    assert len(results) == 1
    assert len(results[0]) == 1
    assert results[0][0].score == pytest.approx(0.8)  # Only one score was provided


def test_multi_entity_consolidation_with_none():
    entities = [
        DummyEntity(name="test", label="LABEL", score=0.8),
        None,
    ]
    strategy = MultiEntityConsolidation(extractor=lambda x: [x] if x else [])
    results = strategy.consolidate(entities, [(0, 2)])
    assert len(results) == 1
    assert len(results[0]) == 1
    assert results[0][0].name == "test"


def test_multi_entity_consolidation_non_frozen():
    entities = [
        NonFrozenEntity(name="test", score=0.8),
        NonFrozenEntity(name="test", score=0.4),
    ]
    strategy = MultiEntityConsolidation(extractor=lambda x: [x])
    results = strategy.consolidate(entities, [(0, 2)])
    assert len(results) == 1
    assert len(results[0]) == 1
    assert results[0][0].name == "test"
    assert results[0][0].score == pytest.approx(0.6)


def test_multi_entity_consolidation_nested():
    dummy1 = DummyEntity(name="inner", label="L")
    entities = [
        NestedEntity(name="outer", details=dummy1, score=0.8),
        NestedEntity(name="outer", details=dummy1, score=0.2),
    ]
    strategy = MultiEntityConsolidation(extractor=lambda x: [x])
    results = strategy.consolidate(entities, [(0, 2)])
    assert len(results) == 1
    assert len(results[0]) == 1
    assert results[0][0].score == pytest.approx(0.5)


def test_single_entity_consolidation_empty():
    strategy = SingleEntityConsolidation(extractor=lambda x: x)
    assert strategy.consolidate([], [(0, 0)]) == [None]


def test_single_entity_consolidation_majority():
    entities = [
        DummyEntity(name="A", label="L", score=0.8),
        DummyEntity(name="A", label="L", score=0.6),
        DummyEntity(name="B", label="L", score=0.9),
    ]
    strategy = SingleEntityConsolidation(extractor=lambda x: x)
    results = strategy.consolidate(entities, [(0, 3)])
    winner = results[0]
    assert winner is not None
    assert winner.name == "A"
    assert winner.score == pytest.approx(0.7)


def test_single_entity_consolidation_tie_break():
    # Tie between A and B, A seen first at index 0
    entities = [
        DummyEntity(name="A", label="L", score=0.8),
        DummyEntity(name="B", label="L", score=0.9),
    ]
    strategy = SingleEntityConsolidation(extractor=lambda x: x)
    results = strategy.consolidate(entities, [(0, 2)])
    winner = results[0]
    assert winner is not None
    assert winner.name == "A"


def test_single_entity_consolidation_none_handling():
    entities = [
        DummyEntity(name="A", label="L", score=0.8),
        None,
        None,
    ]
    strategy = SingleEntityConsolidation(extractor=lambda x: x)
    results = strategy.consolidate(entities, [(0, 3)])
    winner = results[0]
    assert winner is None  # None won by majority (2 vs 1)


def test_single_entity_consolidation_all_none():
    entities = [
        None,
        None,
    ]
    strategy = SingleEntityConsolidation(extractor=lambda x: x)
    results = strategy.consolidate(entities, [(0, 2)])
    winner = results[0]
    assert winner is None



"""Common types."""

from typing import Any

import pydantic


class ModelSettings(pydantic.BaseModel):
    """Settings for model for structured generation.

    :param init_kwargs: kwargs passed on to initialization of structured generator. Not all models use this.
    :param inference_kwargs: kwargs passed on to inference with structured generator.
    :param config_kwargs: Used only if supplied model is a DSPy model object, ignored otherwise. Optional kwargs
        supplied to dspy.configure().
    :param strict: If True, exception is raised if prompt response can't be parsed correctly.
    :param inference_mode: Specifies the inference mode for the model wrapper. If not provided, the model wrapper will
        use its default mode. The available modes depend on the selected model wrapper (e.g., DSPy supports 'predict',
        'chain_of_thought', 'react'; Outlines supports 'text', 'choice', 'regex', 'json').
    """

    init_kwargs: dict[str, Any] | None = None
    inference_kwargs: dict[str, Any] | None = None
    config_kwargs: dict[str, Any] | None = None
    strict: bool = True
    inference_mode: Any | None = None


class TokenUsage(pydantic.BaseModel):
    """Token usage for a model call.

    :param input_tokens: Number of input tokens.
    :param output_tokens: Number of output tokens.
    """

    input_tokens: int | None = None
    output_tokens: int | None = None



"""Pipeline."""

from __future__ import annotations

import copy
import itertools
from collections.abc import Iterable, Iterator, Sized
from pathlib import Path
from typing import Any

import dspy
import tqdm

from sieves.data import Doc
from sieves.serialization import Attribute, Config, Serializable
from sieves.tasks import Task
from sieves.tasks.predictive.evaluation import PipelineEvaluationReport, TaskEvaluationReport


class Pipeline:
    """Pipeline for executing tasks on documents."""

    def __init__(
        self,
        tasks: Iterable[Task] | Task,
        use_cache: bool = True,
    ):
        """Initialize pipeline.

        :param tasks: List of tasks to execute.
        :param use_cache: If True, pipeline will build a cache over processed `Doc`s to ensure that no redundant
            requests will be sent to the model. If False, all `Doc`s will be processed from scratch, regardless of
            whether they have already been processed..
        """
        self._tasks = [tasks] if isinstance(tasks, Task) else list(tasks)
        self._use_cache = use_cache
        self._cache: dict[int, Doc] = {}
        self._cache_stats: dict[str, int] = {"total": 0, "unique": 0, "hits": 0, "misses": 0}
        self._validate_tasks()

    def add_tasks(self, tasks: Iterable[Task]) -> None:
        """Add tasks to pipeline. Revalidates pipeline.

        :param tasks: Tasks to be added.
        """
        self._tasks.extend(tasks)
        self._validate_tasks()

    @property
    def tasks(self) -> list[Task]:
        """Return tasks.

        :return: List of tasks.
        """
        return self._tasks

    @property
    def use_cache(self) -> bool:
        """Return whether pipeline uses cache.

        :return: Whether pipeline uses cache.
        """
        return self._use_cache

    def _validate_tasks(self) -> None:
        """Validate tasks.

        :raises ValueError: On pipeline component signature mismatch.
        """
        task_ids: set[str] = set()

        for i, task in enumerate(self._tasks):
            if task.id in task_ids:
                raise ValueError(f"Task with duplicate ID {task.id}. Ensure unique task IDs.")
            task_ids.add(task.id)

    def _get_unseen_unique_docs(self, docs: Iterable[Doc]) -> Iterable[Doc]:
        """Yield unseen, unique docs.

        I.e. those docs that are not in cache and that are unique within the provided
        collection.

        :param docs: Documents to process.
        """
        doc_hashes: set[int] = set()

        for doc in docs:
            assert doc.text or doc.uri
            doc_cache_id = hash(doc.text or doc.uri)

            if doc_cache_id not in self._cache and doc_cache_id not in doc_hashes:
                doc_hashes.add(doc_cache_id)
                self._cache_stats["unique"] += 1
                yield doc

    def __call__(self, docs: Iterable[Doc], in_place: bool = False, show_progress: bool = True) -> Iterable[Doc]:
        """Process a list of documents through all tasks.

        :param docs: Documents to process.
        :param in_place: Whether to modify documents in-place or create copies.
        :parma show_progress: Whether to show progress bar.
        :return Iterable[Doc]: Processed documents.
        """
        n_docs: int | None = len(docs) if isinstance(docs, Sized) else None
        docs_iters = itertools.tee(docs if in_place else (copy.deepcopy(doc) for doc in docs), 2)
        processed_docs = self._get_unseen_unique_docs(docs_iters[0]) if self._use_cache else docs_iters[0]

        for i, task in enumerate(self._tasks):
            processed_docs = task(processed_docs)

        # If returned docs are not iterators (e.g. returned as lists), get corresponding iterators.
        if not isinstance(processed_docs, Iterator):
            processed_docs = iter(processed_docs)

        # Initialize (nested) progress bar, if progress bar is requested.
        progress_bar: tqdm.tqdm | None = None
        if show_progress:
            progress_bar = tqdm.tqdm(desc="Running pipeline", total=n_docs)

        # Iterate over all docs. Retrieve doc from cache if available, otherwise add to cache.
        for i, doc in enumerate(docs_iters[1]):
            assert doc.text or doc.uri
            self._cache_stats["total"] += 1
            # Docs must either all have URIs or texts. Either is a sufficient identifier. If first task is Ingestion
            # and not all docs have IDs, pipeline fails. If first task is predictive and not all docs have texts,
            # pipeline fails.
            doc_cache_id = hash(doc.text or doc.uri)

            if doc_cache_id not in self._cache:
                # Update cache.
                self._cache_stats["misses"] += 1
                processed_doc = next(processed_docs)
                processed_doc.meta["cached"] = False

                if self._use_cache:
                    self._cache[doc_cache_id] = processed_doc
            else:
                self._cache_stats["hits"] += 1
                processed_doc = copy.deepcopy(self._cache[doc_cache_id])
                processed_doc.meta["cached"] = True

            if show_progress:
                assert progress_bar is not None
                progress_bar.update(1)
                progress_bar.refresh()

            yield processed_doc

        if show_progress:
            assert progress_bar is not None
            progress_bar.close()

    def dump(self, path: Path | str) -> None:
        """Save pipeline config to disk.

        :param path: Target path.
        """
        self.serialize().dump(path)

    def clear_cache(self) -> None:
        """Clear cache."""
        self._cache.clear()
        self._cache_stats = {k: 0 for k in self._cache_stats}

    @classmethod
    def load(cls, path: Path | str, task_kwargs: Iterable[dict[str, Any]]) -> Pipeline:
        """Generate pipeline from disk.

        :param path: Path to config file.
        :param task_kwargs: Values to inject into loaded config.
        :return: Pipeline instance.
        """
        return cls.deserialize(Config.load(path), task_kwargs)

    def serialize(self) -> Config:
        """Serialize pipeline object.

        :return: Serialized pipeline representation.
        """
        return Config.create(
            self.__class__,
            {
                "tasks": Attribute(value=[task.serialize() for task in self._tasks]),
                "use_cache": Attribute(value=self._use_cache),
            },
        )

    @classmethod
    def deserialize(cls, config: Config, tasks_kwargs: Iterable[dict[str, Any]]) -> Pipeline:
        """Generate pipeline from config.

        :param config: Config to generate pipeline from.
        :param tasks_kwargs: Values to inject into task configs. One dict per task (dict can be empty).
        :return: Deserialized pipeline instance.
        """
        config.validate_init_params(cls)
        tasks_kwargs = tuple(tasks_kwargs)

        assert hasattr(config, "tasks")
        assert len(config.tasks.value) == len(tasks_kwargs), ValueError(
            f"len(tasks_kwargs) has to match the number of tasks in this pipeline ({len(config.tasks.value)}."
        )
        assert config.tasks.is_placeholder is False

        # Deserialize tasks.
        tasks: list[Task] = []
        for task_attr, task_kwargs in zip(config.tasks.value, tasks_kwargs):
            # Restore task config, if provided as dict.
            match task_attr:
                case dict():
                    task_config, task_cls = Config.from_dict(task_attr)
                case Config():
                    task_config = task_attr
                    task_cls = task_attr.config_cls
                case _:
                    raise TypeError(f"Deserialization can't handle configs of type {type(task_attr)}.")

            # Deserialize task.
            assert issubclass(task_cls, Serializable)
            assert issubclass(task_cls, Task)
            task = task_cls.deserialize(task_config, **task_kwargs)
            tasks.append(task)

        return cls(tasks=tasks)

    def __getitem__(self, task_id: str) -> Task:
        """Get task with this ID.

        :param task_id: ID of task to fetch.
        :return: Task with specified ID.
        :raises KeyError: If no task with such ID exists.
        """
        for task in self._tasks:
            if task.id == task_id:
                return task

        raise KeyError(f"No task with ID {task_id} exists in this pipeline.")

    def __add__(self, other: Task | Pipeline) -> Pipeline:
        """Chain this pipeline with another task or pipeline using ``+``.

        Returns a new pipeline that executes all tasks of this pipeline first,
        followed by the task(s) provided via ``other``. The original pipeline(s)
        and task(s) are not mutated.

        Cache semantics:
        - The resulting pipeline preserves this pipeline's ``use_cache`` setting
          regardless of whether ``other`` is a task or pipeline.

        :param other: A ``Task`` or another ``Pipeline`` to execute after this pipeline.
        :return: A new ``Pipeline`` representing the chained execution.
        :raises TypeError: If ``other`` is not a ``Task`` or ``Pipeline``.
        """
        if isinstance(other, Pipeline):
            return Pipeline(tasks=[*self._tasks, *other._tasks], use_cache=self._use_cache)

        if isinstance(other, Task):
            return Pipeline(tasks=[*self._tasks, other], use_cache=self._use_cache)

        raise TypeError(f"Cannot chain Pipeline with {type(other).__name__}")

    def __iadd__(self, other: Task | Pipeline) -> Pipeline:
        """Append a task or pipeline to this pipeline in-place using ``+=``.

        Extending with a pipeline appends all tasks from ``other``. Cache setting
        remains unchanged and follows this (left) pipeline.

        Revalidates the pipeline and updates distillation targets.

        :param other: Task or Pipeline to append.
        :return: This pipeline instance (mutated).
        :raises TypeError: If ``other`` is not a ``Task`` or ``Pipeline``.
        """
        if isinstance(other, Task):
            self._tasks.append(other)
        elif isinstance(other, Pipeline):
            self._tasks.extend(other._tasks)
        else:
            raise TypeError(f"Can only add Task or Pipeline to Pipeline with +=, got {type(other).__name__}")
        self._validate_tasks()

        return self

    def evaluate(self, docs: Iterable[Doc], judge: dspy.LM | None = None) -> PipelineEvaluationReport:
        """Evaluate pipeline performance across all tasks.

        :param docs: Documents to evaluate. Must contain gold data for relevant tasks.
        :param judge: Optional judge model for evaluation.
        :return: Pipeline evaluation report.
        """
        reports: dict[str, TaskEvaluationReport[Any]] = {}
        for task in self._tasks:
            docs_iter = itertools.tee(docs, 2)
            try:
                report = task.evaluate(docs_iter[1], judge=judge)
                reports[task.id] = report
            except NotImplementedError:
                # Skip tasks that don't support evaluation.
                continue

        return PipelineEvaluationReport(reports=reports)



"""Schemas for predictive tasks."""

from sieves.tasks.predictive.schemas import (
    classification,
    core,
    information_extraction,
    ner,
    pii_masking,
    question_answering,
    relation_extraction,
    sentiment_analysis,
    summarization,
    translation,
)

__all__ = [
    "classification",
    "core",
    "information_extraction",
    "ner",
    "pii_masking",
    "question_answering",
    "relation_extraction",
    "sentiment_analysis",
    "summarization",
    "translation",
]



# mypy: ignore-errors
import time
from collections.abc import Iterable
import tqdm

import pytest

from sieves import Doc, Pipeline, model_wrappers, tasks
from sieves.tasks import Classification


@pytest.mark.parametrize(
    "batch_runtime",
    [model_wrappers.ModelType.outlines],
    indirect=True,
)
def test_double_task(dummy_docs, batch_runtime) -> None:
    class DummyTask(tasks.Task):
        def _call(self, _docs: Iterable[Doc]) -> Iterable[Doc]:
            _docs = list(_docs)
            for _doc in _docs:
                _doc.results[self._task_id] = "dummy"
            yield from _docs

    pipe = Pipeline(
        [
            DummyTask(task_id="task_1", include_meta=False, batch_size=1),
            DummyTask(task_id="task_2", include_meta=False, batch_size=1),
        ]
    )
    docs = list(pipe(dummy_docs))

    _ = pipe["task_1"]
    with pytest.raises(KeyError):
        _ = pipe["sdfkjs"]

    assert len(docs) == 2
    for doc in docs:
        assert doc.text
        assert doc.results["task_1"]
        assert doc.results["task_2"]
        assert "task_1" in doc.results
        assert "task_2" in doc.results


@pytest.mark.parametrize(
    "batch_runtime",
    [model_wrappers.ModelType.huggingface],
    indirect=True,
)
def test_caching(batch_runtime) -> None:
    labels = ["science", "politics"]
    text_science = (
        "Stars are giant balls of hot gas – mostly hydrogen, with some helium and small amounts of other elements. "
        "Every star has its own life cycle, ranging from a few million to trillions of years, and its properties change"
        " as it ages."
    )
    text_politics = (
        "Politics (from Ancient Greek πολιτικά (politiká) 'affairs of the cities') is the set of activities that are "
        "associated with making decisions in groups, or other forms of power relations among individuals, such as the"
        " distribution of status or resources."
    )

    # Test that uniqueness filtering works.

    n_docs = 10
    docs = [Doc(text=text_science) for _ in range(n_docs)]
    pipe = Pipeline(tasks=Classification(labels=labels, model=batch_runtime.model, model_settings=batch_runtime.model_settings, batch_size=batch_runtime.batch_size))
    docs = list(pipe(docs))
    assert pipe._cache_stats == {"hits": 9, "misses": 1, "total": 10, "unique": 1}
    assert len(docs) == n_docs
    assert docs[0].meta["cached"] is False
    assert all(doc.meta["cached"] is True for doc in docs[1:])

    # Test that uniqueness filtering works while preserving sequence of Docs.

    docs = [Doc(text=text_science), Doc(text=text_politics), Doc(text=text_science)]
    pipe = Pipeline(tasks=Classification(labels=labels, model=batch_runtime.model, model_settings=batch_runtime.model_settings, batch_size=batch_runtime.batch_size))
    docs = list(pipe(docs))
    assert docs[0].text == docs[2].text == text_science
    assert docs[1].text == text_politics
    assert pipe._cache_stats == {"hits": 1, "misses": 2, "total": 3, "unique": 2}
    assert docs[0].meta["cached"] is False
    assert docs[1].meta["cached"] is False
    assert docs[2].meta["cached"] is True

    # Compare uncached with cached mode with identical documents.

    n_docs = 10
    docs = [Doc(text=text_science) for _ in range(n_docs)]
    uncached_pipe = Pipeline(tasks=Classification(labels=labels, model=batch_runtime.model, model_settings=batch_runtime.model_settings, batch_size=batch_runtime.batch_size), use_cache=False)
    cached_pipe = Pipeline(tasks=Classification(labels=labels, model=batch_runtime.model, model_settings=batch_runtime.model_settings, batch_size=batch_runtime.batch_size))

    start = time.time()
    uncached_docs = list(uncached_pipe(docs))
    uncached_time = time.time() - start

    start = time.time()
    cached_docs = list(cached_pipe(docs))
    cached_time = time.time() - start

    assert len(uncached_docs) == len(cached_docs) == n_docs
    assert cached_pipe._cache_stats == {"hits": 9, "misses": 1, "total": 10, "unique": 1}
    assert uncached_pipe._cache_stats == {"hits": 0, "misses": 10, "total": 10, "unique": 0}
    assert uncached_docs[0].meta["cached"] is False
    assert all(doc.meta["cached"] is False for doc in uncached_docs)
    assert cached_docs[0].meta["cached"] is False
    assert all(doc.meta["cached"] is True for doc in cached_docs[1:])
    # Relaxed speed-up requirement: cached pipe should be faster that uncached pipe.
    # This can be a bit flaky, but 3x is usually on the safer side.
    assert cached_time * 3 < uncached_time

    # Test cache reset.
    cached_pipe.clear_cache()
    assert len(cached_pipe._cache) == 0
    assert cached_pipe._cache_stats == {"hits": 0, "misses": 0, "total": 0, "unique": 0}


def test_model_wrapper_imports() -> None:
    """Tests direct runtime imports."""
    from sieves.model_wrappers import DSPy, GliNER, HuggingFace, LangChain, Outlines  # noqa: F401


def test_add_task_task(dummy_docs) -> None:
    """Chaining two tasks with ``+`` yields a working Pipeline with both results."""

    class DummyTask(tasks.Task):
        def _call(self, _docs: Iterable[Doc]) -> Iterable[Doc]:
            _docs = list(_docs)
            for _doc in _docs:
                _doc.results[self._task_id] = "ok"
            yield from _docs

    pipe = (
        DummyTask(task_id="t1", include_meta=False, batch_size=-1)
        + DummyTask(task_id="t2", include_meta=False, batch_size=-1)
    )

    assert isinstance(pipe, Pipeline)
    docs = list(pipe(dummy_docs))
    assert len(docs) == 2
    for d in docs:
        assert d.results["t1"] == "ok"
        assert d.results["t2"] == "ok"


def test_add_pipeline_task_and_task_pipeline(dummy_docs) -> None:
    """Chaining Pipeline+Task and Task+Pipeline produces identical outputs order-wise."""

    class DummyTask(tasks.Task):
        def _call(self, _docs: Iterable[Doc]) -> Iterable[Doc]:
            _docs = list(_docs)
            import time
            time.sleep(3)
            for _doc in _docs:
                _doc.results[self._task_id] = "ok"
            yield from _docs

    t1 = DummyTask(task_id="t1", include_meta=False, batch_size=1)
    t2 = DummyTask(task_id="t2", include_meta=False, batch_size=1)

    p1 = Pipeline([t1])
    p2 = p1 + t2
    p3 = t1 + Pipeline([t2])

    for p in (p2, p3):
        docs = list(p(dummy_docs))
        assert len(docs) == 2
        for d in docs:
            assert d.results["t1"] == "ok"
            assert d.results["t2"] == "ok"


if __name__ == '__main__':

    class DummyTask(tasks.Task):
        def _call(self, _docs: Iterable[Doc]) -> Iterable[Doc]:
            _docs = list(_docs)
            import time

            for _doc in _docs:
                time.sleep(.5)
                _doc.results[self._task_id] = "ok"
                yield _doc

    t1 = DummyTask(task_id="t1", include_meta=False, batch_size=1)
    t2 = DummyTask(task_id="t2", include_meta=False, batch_size=1)
    pipe = t1 + t2

    dummy_docs = [Doc(text=str(i)) for i in (1, 2, 3)]
    docs = list(pipe(dummy_docs))
    assert len(docs) == 3

    dummy_docs = [Doc(text=str(i)) for i in (4, 5, 6)]
    for i in tqdm.tqdm([1, 2, 3], desc="Outer loop", total=3, position=0, leave=False):
        docs = list(pipe(dummy_docs))
        assert len(docs) == 3

    dummy_docs = [Doc(text=str(i)) for i in (7, 8, 9)]
    docs = list(pipe(dummy_docs, show_progress=False))
    assert len(docs) == 3


def test_add_pipeline_pipeline(dummy_docs) -> None:
    """Chaining Pipeline+Pipeline concatenates tasks and preserves left cache semantics."""

    class DummyTask(tasks.Task):
        def _call(self, _docs: Iterable[Doc]) -> Iterable[Doc]:
            _docs = list(_docs)
            for _doc in _docs:
                _doc.results[self._task_id] = "ok"
            yield from _docs

    p_left = Pipeline([DummyTask(task_id="left", include_meta=False, batch_size=-1)])
    p_right = Pipeline([DummyTask(task_id="right", include_meta=False, batch_size=-1)])

    p = p_left + p_right
    docs = list(p(dummy_docs))
    for d in docs:
        assert d.results["left"] == "ok"
        assert d.results["right"] == "ok"


def test_add_does_not_mutate_originals() -> None:
    """Chaining should not mutate the original Task or Pipeline instances."""

    class DummyTask(tasks.Task):
        def _call(self, _docs: Iterable[Doc]) -> Iterable[Doc]:
            yield from _docs

    t1 = DummyTask(task_id="t1", include_meta=False, batch_size=-1)
    t2 = DummyTask(task_id="t2", include_meta=False, batch_size=-1)

    p1 = Pipeline([t1])
    p2 = Pipeline([t2])
    p3 = p1 + p2

    assert len(p1._tasks) == 1
    assert len(p2._tasks) == 1
    assert len(p3._tasks) == 2
    assert p3._tasks[0] is t1 and p3._tasks[1] is t2


def test_add_cache_semantics(dummy_docs) -> None:
    """Verify cache propagation rules for all supported chaining combinations."""

    class DummyTask(tasks.Task):
        def _call(self, _docs: Iterable[Doc]) -> Iterable[Doc]:
            yield from _docs

    t1 = DummyTask(task_id="t1", include_meta=False, batch_size=-1)
    t2 = DummyTask(task_id="t2", include_meta=False, batch_size=-1)

    p_uncached = Pipeline([t1], use_cache=False)
    p_cached = Pipeline([t2], use_cache=True)

    # Left pipeline wins
    p = p_uncached + p_cached
    assert p._use_cache is False

    p = p_uncached + t2
    assert p._use_cache is False

    # Task + Pipeline adopts right pipeline cache
    p = t1 + p_cached
    assert p._use_cache is True

    # Task + Task defaults to True
    p = t1 + t2
    assert p._use_cache is True


def test_iadd_pipeline_task(dummy_docs) -> None:
    """Pipeline ``+= Task`` appends in-place and preserves order and cache semantics."""

    class DummyTask(tasks.Task):
        def _call(self, _docs: Iterable[Doc]) -> Iterable[Doc]:
            _docs = list(_docs)
            for _doc in _docs:
                _doc.results[self._task_id] = "ok"
            yield from _docs

    t1 = DummyTask(task_id="t1", include_meta=False, batch_size=-1)
    t2 = DummyTask(task_id="t2", include_meta=False, batch_size=-1)

    p = Pipeline([t1], use_cache=False)
    p += t2

    # Mutated in place
    assert len(p.tasks) == 2
    assert p.tasks[0] is t1 and p.tasks[1] is t2
    assert p.use_cache is False

    docs = list(p(dummy_docs))
    assert len(docs) == 2
    for d in docs:
        assert d.results["t1"] == "ok"
        assert d.results["t2"] == "ok"


def test_iadd_pipeline_pipeline(dummy_docs) -> None:
    """Pipeline ``+= Pipeline`` appends all tasks and preserves left cache semantics."""

    class DummyTask(tasks.Task):
        def _call(self, _docs: Iterable[Doc]) -> Iterable[Doc]:
            _docs = list(_docs)
            for _doc in _docs:
                _doc.results[self._task_id] = "ok"
            yield from _docs

    left = Pipeline([DummyTask(task_id="left", include_meta=False, batch_size=-1)], use_cache=False)
    right = Pipeline([DummyTask(task_id="right", include_meta=False, batch_size=-1)], use_cache=True)

    left += right
    assert len(left.tasks) == 2
    assert left.tasks[0].id == "left" and left.tasks[1].id == "right"
    # Left cache semantics preserved
    assert left.use_cache is False

    docs = list(left(dummy_docs))
    assert len(docs) == 2
    for d in docs:
        assert d.results["left"] == "ok"
        assert d.results["right"] == "ok"



"""Import 3rd-party libraries required for distillation.

If library can't be found, placeholder model wrappers is imported instead.

This allows us to import everything downstream without having to worry about optional dependencies. If a user specifies
a non-installed distillation framework, we terminate with an error.
"""

# mypy: disable-error-code="no-redef"

try:
    import sentence_transformers
except (ModuleNotFoundError, ImportError):
    sentence_transformers = None


try:
    import setfit
except (ModuleNotFoundError, ImportError):
    setfit = None


try:
    import model2vec
    import model2vec.train
except (ModuleNotFoundError, ImportError):
    model2vec = None


__all__ = ["model2vec", "sentence_transformers", "setfit"]



"""Fallback ingestion task when optional dependencies are unavailable."""

from collections.abc import Iterable
from typing import Any, override

from sieves.data import Doc
from sieves.tasks.core import Task

Converter = Any


class MissingIngestion(Task):
    """Placeholder task raised when ingestion backends are missing.

    Instantiating this class informs users to install the `ingestion` extra.
    """

    @override
    def __init__(self, *args: Any, **kwargs: Any) -> None:  # noqa: D401
        raise ImportError("Optional ingestion dependency not installed. Install with: uv sync --extra ingestion")

    @override
    def _call(self, docs: Iterable[Doc]) -> Iterable[Doc]:
        raise NotImplementedError



"""Sieves."""

import sieves.tasks as tasks
from sieves.data import Doc
from sieves.model_wrappers import ModelSettings
from sieves.pipeline import Pipeline

__all__ = ["Doc", "ModelSettings", "tasks", "Pipeline"]



"""Chunking task."""

from __future__ import annotations

import itertools
from collections.abc import Callable, Iterable
from typing import Any

import chonkie

from sieves.data.doc import Doc
from sieves.serialization import Config
from sieves.tasks.core import Task
from sieves.tasks.preprocessing import chunking
from sieves.tasks.preprocessing.chunking import chonkie_, naive

_ChunkerArgType = chonkie.BaseChunker | int
_ChunkerType = chonkie_.Chonkie | naive.NaiveChunker


class Chunking(Task):
    """Task for chunking documents using different strategies.

    This task acts as a wrapper around specific chunker implementations,
    allowing for flexible configuration based on the provided chunker object or interval.
    """

    def __init__(
        self,
        chunker: _ChunkerArgType,
        task_id: str | None = None,
        include_meta: bool = False,
        batch_size: int = -1,
        condition: Callable[[Doc], bool] | None = None,
    ):
        """Initialize the Chunker task.

        :param chunker: The chunker instance (chonkie.BaseChunker) or the interval (int) for NaiveChunker.
        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for processing. Use -1 to process all documents at once.
        :param condition: Optional callable that determines whether to process each document.
        """
        super().__init__(task_id=task_id, include_meta=include_meta, batch_size=batch_size, condition=condition)
        self._chunker_arg = chunker
        self._task = self._init_chunker_task()

    def _init_chunker_task(self) -> _ChunkerType:
        """Initialize the specific chunker task based on the type of _chunker_arg.

        :return: Initialized chunker task instance.
        :raises TypeError: If the type of _chunker_arg is not supported.
        """
        chunker_task: _ChunkerType

        match self._chunker_arg:
            case chunker if isinstance(chunker, chonkie.BaseChunker):
                chunker_task = chunking.chonkie_.Chonkie(
                    chunker=chunker,
                    task_id=self.id,
                    include_meta=self._include_meta,
                    batch_size=self._batch_size,
                )
            case interval if isinstance(interval, int):
                chunker_task = chunking.naive.NaiveChunker(
                    interval=interval,
                    task_id=self.id,
                    include_meta=self._include_meta,
                    batch_size=self._batch_size,
                )
            case _:
                raise TypeError(
                    f"Unsupported type for 'chunker' argument: {type(self._chunker_arg)}. "
                    f"Expected chonkie.BaseChunker or int."
                )

        return chunker_task

    def _call(self, docs: Iterable[Doc]) -> Iterable[Doc]:
        """Process documents by chunking their text.

        :param docs: Documents to process.
        :return: Processed documents with chunks added.
        """
        docs_iters = itertools.tee(docs, 2)
        assert all(doc.text for doc in docs_iters[0]), ValueError("Documents have to have a value for .text.")
        yield from self._task(docs_iters[1])

    @property
    def _state(self) -> dict[str, Any]:
        """Return attributes to serialize.

        :return: Dict of attributes to serialize.
        """
        return {
            **super()._state,
            "chunker": self._chunker_arg,
        }

    @classmethod
    def deserialize(cls, config: Config, **kwargs: dict[str, Any]) -> Chunking:
        """Generate Chunker instance from config.

        :param config: Config to generate instance from.
        :param kwargs: Values to inject into loaded config.
        :return: Deserialized Chunker instance.
        """
        return cls(**config.to_init_dict(cls, **kwargs))



"""Schemas for translation task."""

from __future__ import annotations

import dspy
import pydantic

from sieves.model_wrappers import dspy_, langchain_, outlines_
from sieves.tasks.predictive.schemas.core import FewshotExample as BaseFewshotExample


class FewshotExample(BaseFewshotExample):
    """Example for translation few-shot prompting.

    Attributes:
        text: Input text.
        translation: Translated text.
        score: Confidence score.
    """

    translation: str
    score: float | None = None

    @property
    def target_fields(self) -> tuple[str, ...]:
        """Return target fields.

        :return: Target fields.
        """
        return ("translation",)


# --8<-- [start:Result]
class Result(pydantic.BaseModel):
    """Result of a translation task. Contains the translated text and a confidence score.

    Attributes:
        translation: Translated text.
        score: Confidence score.
    """

    translation: str = pydantic.Field(description="The input text translated into the target language.")
    score: float | None = pydantic.Field(
        default=None, description="Provide a confidence score for the translation, between 0 and 1."
    )


# --8<-- [end:Result]


TaskModel = dspy_.Model | langchain_.Model | outlines_.Model
TaskPromptSignature = type[dspy.Signature] | type[pydantic.BaseModel]
TaskResult = Result



"""
Tests for model wrapper usage documentation snippets.
"""
import os
import pytest
from sieves import tasks

def test_dspy_usage():
    # --8<-- [start:dspy-usage]
    import dspy
    from sieves import tasks

    # Initialize a DSPy Language Model
    model = dspy.LM("openai/gpt-4o-mini", api_key="dummy")

    # Pass it to a task
    task = tasks.SentimentAnalysis(model=model)
    # --8<-- [end:dspy-usage]
    assert task

def test_outlines_usage():
    # --8<-- [start:outlines-usage]
    import outlines
    from transformers import AutoModelForCausalLM, AutoTokenizer
    from sieves import tasks

    # Initialize an Outlines model
    model_name = "HuggingFaceTB/SmolLM2-135M-Instruct"
    model = outlines.models.from_transformers(
        AutoModelForCausalLM.from_pretrained(model_name),
        AutoTokenizer.from_pretrained(model_name)
    )

    # Pass it to a task
    task = tasks.SentimentAnalysis(model=model)
    # --8<-- [end:outlines-usage]
    assert task

def test_langchain_usage():
    # --8<-- [start:langchain-usage]
    from langchain_openai import ChatOpenAI
    from sieves import tasks

    # Initialize a LangChain Chat Model
    model = ChatOpenAI(model="gpt-4o-mini", api_key="dummy")

    # Pass it to a task
    task = tasks.SentimentAnalysis(model=model)
    # --8<-- [end:langchain-usage]
    assert task

def test_gliner_usage():
    # --8<-- [start:gliner-usage]
    import gliner2
    from sieves import tasks

    # Initialize a GLiNER model
    model = gliner2.GLiNER2.from_pretrained("fastino/gliner2-base-v1")

    # Pass it to a task (Note: GLiNER is specialized for extraction/classification)
    task = tasks.NER(
        entities=["PERSON", "LOC"],
        model=model
    )
    # --8<-- [end:gliner-usage]
    assert task

def test_huggingface_usage():
    # --8<-- [start:huggingface-usage]
    import transformers
    from sieves import tasks

    # Initialize a Hugging Face pipeline
    model = transformers.pipeline(
        "zero-shot-classification",
        model="MoritzLaurer/xtremedistil-l6-h256-zeroshot-v1.1-all-33"
    )

    # Pass it to a task
    task = tasks.Classification(
        labels=["positive", "negative"],
        model=model
    )
    # --8<-- [end:huggingface-usage]
    assert task
